#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2020"
__version__		= "0.0.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyGetGTFSources.py
#
#
#	Copyright (c) Sander Granneman 2020
#
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import time
from collections import defaultdict
from optparse import *

def getGTFSourceList(file_object):
	"""Looks at the second column in the GTF file and stores the information in a set list,
	hence, only unique annotations/features are stored. These will be printed to the standard output"""
	sourcelist = set()
	count	   = defaultdict(int)
	for line in file_object:
		if line[0] != "#":
			try:
				Fld = line.strip("\n").split("\t")
				sourcelist.add(Fld[1])
				count[Fld[1]] += 1
			except IndexError:
				sys.stderr.write("IndexError at line:\n%s" % (line))
		else:
			continue
	sortedsourcelist = list()
	sortedsourcelist.extend(sourcelist)
	return sortedsourcelist,count

def main():
	parser = OptionParser(usage="usage: %prog [options] --gtf=myfavgtf --count", version="%s" % __version__)
	parser.add_option("--gtf",dest="gtf_file",metavar="Yourfavoritegtf.gtf",help="type the path to the gtf file that you want to use. By default it expects data from the standard input",default=None)
	parser.add_option("-o","--outfile",dest="outfile",help="type the name and path of the file you want to write the output to. Default is standard output",default=None)
	parser.add_option("--count",dest="count",action="store_true",help="with this flag you the program will count the occurence for each source/annotation in the gtf file",default=False)
	(options, args) = parser.parse_args()
	data = sys.stdin
	out	 = sys.stdout
	if options.outfile:
		out = open(options.outfile,"w")
	if options.gtf_file:
		data = open(options.gtf_file,"r")
	sources,counts = getGTFSourceList(data)
	out.write("# %s\n# %s\n# source list generated from: %s\n" % (' '.join(sys.argv),time.ctime(),options.gtf_file))
	if options.count:
		for i in sources:
			out.write("%s\t%s\n" % (i,counts[i]))
	else:
		for i in sources:
			out.write("%s\n" % (i))

if __name__ == "__main__":
	main()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "production"

##################################################################################
#
#	pyCalculateChromosomeLengths.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
from optparse import *
from pyCRAC.Classes.Exceptions import FileTypeError

def main():
	""" this program converts fasta files to the tabular output """
	parser = OptionParser(usage="usage: %prog -f filename.fasta", version="%s" % __version__)
	parser.add_option("-f", "--input_file",dest="input_file",metavar="chromosomes.fasta",help="provide the name and path of your fasta or tab genomic sequence file. Default is standard input.",default=None)
	parser.add_option("--file_type",choices=["fasta","tab"],metavar="fasta",help="provide the file type (fasta or tab). Default is fasta",default="fasta")
	parser.add_option("-o","--output_file",dest="output_file",metavar="FILE",help="to provide an output file name. Default is standard output",default=None)
	parser.add_option("--trim",dest="trim",action="store_true",help="to trim the header of the chromosome name. This shortens the name to make it the same as in the GTF file. Please double check the output to make sure this went correctly!")
	(options, args) = parser.parse_args()
	input  = sys.stdin
	output = sys.stdout
	if options.input_file:
		input  = open(options.input_file,"r")
	if options.output_file:
		output = open(options.output_file,"w")
	chromosome = str()
	sequence   = str()
	if options.file_type == "fasta":
		for line in input:
			if line.startswith(">") and sequence:		# if a new id has been found and sequence exists, print to file
				output.write("%s\t%s\n" % (chromosome,len(sequence)))
				if options.trim:
					chromosome = line[1:].strip().split()[0]  # only retain the first part!
				else:
					chromosome = line[1:].strip()
				sequence = str()
			elif line.startswith(">"):					# if only a new id has been found, store the id
				if options.trim:
					chromosome = line[1:].strip().split()[0]  # only retain the first part!
				else:
					chromosome = line[1:].strip()
			else:
				sequence += line.strip()				# else, concatenate the sequence strings
		output.write("%s\t%s\n" % (chromosome,len(sequence)))
	elif options.file_type == "tab":
		for line in input:
			chromosome,sequence = line.strip().split()
			output.write("%s\t%s\n" % (chromosome,len(sequence)))
		
	else:
		raise FileTypeError("\nI do not recognize the %s format. Please choose from 'fasta' or 'tab'\n")
	output.close()
		
if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__      = "Sander Granneman"
__copyright__   = "Copyright 2019"
__version__     = "0.0.2"
__credits__     = ["Sander Granneman"]
__maintainer__  = "Sander Granneman"
__email__       = "sgrannem@staffmail.ed.ac.uk"
__status__      = "Production"

##################################################################################
#
#   Exceptions.py
#
#
#   Copyright (c) Sander Granneman 2019
#   
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#   
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#   
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#
##################################################################################

class AnnotationError(Exception):
    """ Called when an annotation could not be found in the GTF annotation file """
    pass

class SourceError(Exception):
    """ Called when the user specifies a source (i.e. CDS,exon,5UTR, etc) that does not exist """
    pass

class LookupError(Exception):
    """ Called when a variable could not be found in an array or dictionary """
    pass
    
class NoSequenceError(Exception):
    """ Called when no sequence information is available. Usually because the genomic reference sequence has not been loaded """
    pass
    
class InputError(Exception):
    """ Called when an additional input is required to run the function/method or when the user uses an incorrect input """
    pass
    
class BaseQualityError(Exception):
    """ Called when a base quality could not be calculated """
    pass
    
class FileTypeError(Exception):
    """ Called the program could not recognise the file type of the input file or when the file type is not supported """
    pass
    
class FormatError(Exception):
    """ Called when a file or sequence format could not be recognised """
    pass
    
class ProcessingError(Exception):
    """ Called when a specific method needs to be run or a file needs to be processed first in order to continue """
    pass
    
class NoResultsError(Exception):
    """ Called when the method or function did not return any results """
    pass
    
class ClusteringError(Exception):
    """ Called during a Clustering error """
    pass


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.1.4"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	Pileups
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import numpy as np
import time
from collections import defaultdict
from pyCRAC.Parsers import ParseAlignments
from pyCRAC.Classes.Exceptions import *
from pyCRAC.Methods import strand_converter,reverse_strand,getfilename

class Pileup():
	"""Generates tabular pileup files for novo, sam or pyReadCounters/pyMotif GTF files.
	Requires a GTF annotation file, tab delimited genomic reference 
	sequence and alignment file"""
	def __init__(self,data_file,gtf,genes,file_type="novo",sequence="genomic",muts_filter=None,debug=False):
		self.substitutions		= defaultdict(int)
		self.deletions			= defaultdict(int)
		self.gene_hits			= defaultdict(int)
		self.hits				= defaultdict(int)
		self.reads_in_cluster	= defaultdict(list)
		self.data				= None
		self.genes				= genes
		self.file_type			= file_type.lower()
		self.file_name			= data_file
		self.file_list			= list()
		self.orientation		= "sense"
		self.__muts_filter		= muts_filter
		self.__gtf				= gtf
		self.__gene_sequence	= defaultdict(str)
		self.__seq_line			= defaultdict(str)
		self.__itercoordinates	= defaultdict(list)
		self.__sequence			= sequence
		self.__supported_filetypes = ["novo","sam","gtf"] # index = [0,1,2]; [:2] = novo and sam, [2] = gtf .
		self.__determineFileType(data_file,file_type,debug)
		self.__orientations = ["sense","anti-sense"]

		for gene in self.genes:						   
			if self.__sequence == "coding":
				self.__gene_sequence[gene] = gtf.codingSequence(gene)
			elif self.__sequence == "genomic":
				self.__gene_sequence[gene] = gtf.genomicSequence(gene)
			self.__itercoordinates[gene] = gtf.geneIterCoordinates(gene,coordinates=sequence,output="numpy")
			self.__seq_line[gene] = ">%s\n%s" % (gene,self.__gene_sequence[gene])
			
	def __determineFileType(self,data_file,file_type,debug):
		"""Checks the file type of the data file"""
		if self.file_type in self.__supported_filetypes[:2]:
			self.data = ParseAlignments.ParseAlignmentFile(data_file,file_type,debug=debug)
			self.data.gtfdatatype = None
		elif self.file_type == self.__supported_filetypes[2]:
			self.data = ParseAlignments.ParseCountersOutput(data_file)
		else:
			raise FileTypeError("\n\nThe file type % is not supported by pyPileup, please choose from the following options: %s\n" % (file_type, ",".join(self.__supported_filetypes))) 
	
#--------------------------------------------- Processing alignment files ---------------------------------------------------

	def countGeneNucleotideCoverage(self,unique=False,distance=1000,length=1000,limit=-1,max_reads=-1,align_quality=0,align_score=0,min_overlap=1,blocks=False,ignorestrand=False,fiveends=False,threeends=False,orientation="sense"):
		""" Takes the alignment file read sequences and aligns them to selected reference sequences"""	
		assert orientation in self.__orientations, "Error! Orientation %s is not recognized. Please choose %s" % (orientation, " or ".join(self.__orientations))
		self.orientation = orientation
		self.data.number_of_reads = 1
		self.__align_quality = align_quality
		self.__align_score	 = align_score
		self.__length		 = length
		self.__blocksremoved = blocks
		overlap = list()

#------------------------------------ if the input file is a novo or sam file -----------------------------------------------
			
		if self.file_type in self.__supported_filetypes[:2]:
			while self.data.getRead(align_quality=align_quality,align_score=align_score,maximum=max_reads,blocks=blocks,distance=distance,length=length,unique=unique,muts_filter=self.__muts_filter):		# novo or sam input file
				if not self.genes:
					break
				strand = strand_converter(self.data.strand)
				if orientation == "anti-sense":
					strand = reverse_strand(strand)
				for gene in self.genes:
					if self.data.chromosome == self.__gtf.chromosome(gene) and strand == self.__gtf.strand(gene):
						overlap = list()
						if fiveends:				# will report the first nucleotide upstream of the
							if self.__gtf.strand(gene) == "+":
								if self.data.read_start in self.__itercoordinates[gene]:
									overlap = [self.data.read_start]
							else:
								if self.data.read_end in self.__itercoordinates[gene]:
									overlap = [self.data.read_end-1]	
							min_overlap = 1
						elif threeends:
							if self.__gtf.strand(gene) == "+":	  
								if self.data.read_end in self.__itercoordinates[gene]:
									overlap = [self.data.read_end-1]
							else:
								if self.data.read_start in self.__itercoordinates[gene]:
									overlap = [self.data.read_start]
							min_overlap = 1
						else:
							overlap = list(self.__itercoordinates[gene][(self.__itercoordinates[gene] >= self.data.read_start) & (self.__itercoordinates[gene] < self.data.read_end)])
						if len(overlap) >= min_overlap:
							if self.gene_hits[gene] == limit:
								self.genes.remove(gene)
							else:
								for i in overlap:
									self.hits[gene,i] += self.data.number_of_reads
								for j in self.data.substitutions:
									self.substitutions[gene,j] += self.data.number_of_reads
								for k in self.data.deletions:
									self.deletions[gene,k] += self.data.number_of_reads
								self.gene_hits[gene] += self.data.number_of_reads

#------------------------------------ if the input file is a pyReadCounter/pyMotif/pyClusterReads GTF file -----------------------------------

		elif self.file_type in self.__supported_filetypes[2]:
			while self.data.readLineByLine():
				if not self.genes:
					break
				strand = strand_converter(self.data.strand)
				if orientation == "anti-sense":
					strand = reverse_strand(strand)
				genesinboth = set(self.genes) & set(self.data.genes)
				if genesinboth:
					for gene in genesinboth:
						if self.data.chromosome == self.__gtf.chromosome(gene) and strand == self.__gtf.strand(gene):
							if fiveends:				# will report the first nucleotide upstream of the
								if self.__gtf.strand(gene) == "+":
									if self.data.read_start in self.__itercoordinates[gene]:
										overlap = [self.data.read_start]
								else:
									if self.data.read_end in self.__itercoordinates[gene]:
										overlap = [self.data.read_end-1]	
								min_overlap = 1
							elif threeends:
								if self.__gtf.strand(gene) == "+":	  
									if self.data.read_end in self.__itercoordinates[gene]:
										overlap = [self.data.read_end-1]
								else:
									if self.data.read_start in self.__itercoordinates[gene]:
										overlap = [self.data.read_start]
								min_overlap = 1
							else:
								overlap = list(self.__itercoordinates[gene][(self.__itercoordinates[gene] >= self.data.read_start) & (self.__itercoordinates[gene] < self.data.read_end)])
							if len(overlap) >= min_overlap:
								for i in overlap:
									self.hits[gene,i] += self.data.number_of_reads
								for j in self.data.substitutions:
									self.substitutions[gene,j] += self.data.number_of_reads
								for k in self.data.deletions:
									self.deletions[gene,k] += self.data.number_of_reads
								self.gene_hits[gene] += self.data.number_of_reads

#-------------------------------------------------------- Printing to output files --------------------------------------------------------

	def __createFileHandles(self,gene,file_type,file_extension="pileup"):
		""" does what it says """
		filename = str()							  
		filename = "%s-%s_%s_%s_%s.%s" % (file_type,self.__insert,gene,self.__sequence,getfilename(self.file_name),file_extension)
		self.file_list.append(filename)
		return filename

	def printPileup(self,out_file=None):
		""" does what it says """
		if not self.gene_hits:
			raise NoResultsError("\n\nNo hits were found. Are you sure you used the right options?\nDid you set the file type option (novo,sam,gtf)?\nDid you include the correct GTF file?\nDid you include the correct tab file?\nAre the chromosome names in the novo/SAM file the same as in the GTF reference file?\n")
		file_out  = defaultdict(str)
		self.__insert = "reads"
		if self.data.gtfdatatype:
			if self.data.gtfdatatype == "blocks":
				self.__insert = "cDNAs"
			elif self.data.gtfdatatype == "clusters":
				self.__insert = "clusters"
			elif self.data.gtfdatatype == "peaks":
				self.__insert = "peaks"
		else:
			if self.__blocksremoved:
				self.__insert = "cDNAs"
		if out_file:						## in Galaxy all the pileups are printed into a single file, whereas in the command line sense and anti-sense hits are printed in seperate files
			file_name = str()
			if out_file == "default":
				file_name = open(self.__createFileHandles("genes","all"),"w")
			else:
				file_name = open(out_file, "w")
			for gene in sorted(self.gene_hits):
				file_out[gene] = file_name
		else:
			for gene in sorted(self.gene_hits):
				file_out[gene] = open(self.__createFileHandles(gene,self.orientation),"w")
		for gene in sorted(self.gene_hits):
			if self.__gtf.strand(gene) == "-":
				self.__itercoordinates[gene] = self.__itercoordinates[gene][::-1]
			file_out[gene].write("# %s\n" % (' '.join(sys.argv)))
			file_out[gene].write("# %s\n" % (time.ctime()))
			file_out[gene].write("# align quality threshold:\t%s\n# align score threshold:\t%s\n# read length threshold:\t%s\n" % (self.__align_quality,self.__align_score,self.__length))
			file_out[gene].write("# total number of %s:\t%s\n" % (self.__insert,self.data.total_reads))
			file_out[gene].write("# total number of mapped %s:\t%s\n" % (self.__insert,self.data.mapped_reads))
			file_out[gene].write("# total number paired reads:\t%s\n" % (self.data.paired_reads))
			file_out[gene].write("# total number single reads:\t%s\n" % (self.data.single_reads))
			file_out[gene].write("# total %s %s %s found:\t%s\n" % (self.orientation,gene,self.__insert,self.gene_hits[gene]))
			file_out[gene].write("# gene\tposition\tnucleotide\thits\tsubstitutions\tdeletions\n")
			try:
				for rank,numbers in enumerate(self.__itercoordinates[gene],start=1):
					file_out[gene].write("%s\t%s\t%s\t%d\t%d\t%d\n" % (gene, rank, self.__gene_sequence[gene][rank-1], self.hits[gene,numbers],self.substitutions[gene,numbers],self.deletions[gene,numbers]))
			except IndexError:
				sys.stderr.write("IndexError for gene %s with gene length %s at position %s\n" % (gene,len(self.__gene_sequence[gene]),rank-1))
				pass
			file_out[gene].write("\n\n")
		for i in file_out:
			file_out[i].close()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.5.7"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyBinCollector.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import os
import sys
import time
import tempfile
import shutil
import numpy as np
from pyCRAC.Classes.Coverage import BinCollector
from pyCRAC.Parsers import GTF2
from pyCRAC.Methods import splitData,getfilename,sortbylistlength
from optparse import *
from collections import defaultdict
							
def main():
	parser = OptionParser(usage="usage: %prog [options] -f filename --gtf=yeast.gtf -n 100 -r 50 --outputall --normalize", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("-f", "--input_file", dest="input_file",help="Provide the path and name of the pyClusterReads, pyReadCounters or pyMotif GTF output file. By default the program expects data from the standard input.", metavar="FILE")
	files.add_option("-o","--output_file",dest="output_file",help="To set an output file name. Do not add a file extension. By default, if the --outputall flag is not used, the program writes to the standard output.",default=None)
	files.add_option("--gtf",dest="annotation_file",metavar="yeast.gtf",help="type the path to the gtf annotation file that you want to use.")
	common = OptionGroup(parser, "Common pyCRAC options")
	common.add_option("-r","--range",dest="range",type="int",metavar="100",help="allows you to set the length of the UTR regions. If you set '-r 50' or '--range=50', then the program will set a fixed length (50 bp) regardless of whether the GTF file has genes with annotated UTRs.",default=0)
	common.add_option("-s","--sequence",dest="sequence",metavar="intron",choices=["intron","coding","genomic","CDS","exon","5UTR","3UTR","5ss","3ss","TSS","5end","3end","CDSstart","CDSend"],help="with this option you can select whether you want to generate bins from the coding or genomic sequence or intron, exon, CDS, UTR, 5ss, 3ss, TSS, 5end, 3end, CDSstart or CDSend coordinates. Default = genomic.",default="genomic")
	common.add_option("--ignorestrand",dest="nostrand",action="store_true",help="To ignore strand information and all reads overlapping with genomic features will be considered sense reads. Useful for analysing ChIP or RIP data.",default=False)
	common.add_option("-v","--verbose",dest="verbose",action="store_true",help="prints all the status messages to a file rather than the standard output.",default=False)
	common.add_option("-a","--annotation",dest="annotation",help="select an annotation (i.e. protein_coding, ncRNA, sRNA, rRNA, tRNA, snoRNA) you would like to focus your search on. Default = all.", metavar="protein_coding",default=None)
	common.add_option("--readoverlap",dest="readoverlap",type="int",metavar="1",help="the minimum number of nucleotides a read has to overlap with a feature to be considered a hit. Default is one nucleotide.",default=1)
	specific = OptionGroup(parser, "pyBinCollector specific options")
	specific.add_option("--anti_sense",dest="anti_sense",action="store_true",help="by default pyBinCollector only looks at intervals that overlap 'sense' with features. Use this flag if you only want to plot intervals that map anti-sense.",default=False)
	specific.add_option("--deletions",dest="deletions",action="store_true",help="By default pyBinCollector only looks at interval coordinates. Use this flag to make distribution plots of deletions over genomic features.",default=False)
	specific.add_option("--substitutions",dest="substitutions",action="store_true",help="By default pyBinCollector only looks at interval coordinates. Use this flag to make distribution plots of substitutions over genomic features.",default=False)	
	specific.add_option("--min_length",dest="min_length",metavar="20",type="int",help="to set a minimum length threshold for genes. Genes shorter than the minimal length will be discarded. Default = 1.", default=1)
	specific.add_option("--max_length",dest="max_length",metavar="10000",type="int",help="to set a maximum length threshold for genes. Genes larger than the maximum length will be discarded. Default = 100000000.", default=100000000)
	specific.add_option("-n","--numberofbins",dest="numberofbins",help="Use this flag if you want to normalize feature lengths by dividing them in an equal number fo bins. Default is None.",metavar="100",type="int", default=0)
	specific.add_option("--binoverlap",dest="binoverlap",nargs=2,type="int",metavar="2 4",help="Allows you to print out intervals that were allocated to specific bins. This option expects two numbers, one for each bin, separated by a space. For example: --binoverlap 20 30.", default=[])
	specific.add_option("--overlap",dest="overlap",action="store_true",help="bedtool-like function. Allows you to print out overlap between intervals and specific features such as 5UTR,3UTR,exons, etc without binning.", default=False)
	specific.add_option("--outputall",dest="outputall",action="store_true",help="use this flag to output the interval distribution for each individual gene, rather than making a cumulative coverage plot. Useful for making box plots or for making heat maps.", default=False)
	specific.add_option("--normalize",dest="normalize",action="store_true",help="calculates nucleotide density frequencies for each bin for each feature by dividing the value in each bin by the total read density for the feature.", default=False)
	specific.add_option("--unique",dest="unique",action="store_true",help="To only count an interval in the data gtf file once. Essentially ignores the number in column 5 of the pyCRAC gtf files.", default=False)
	parser.add_option_group(files)
	parser.add_option_group(common)
	parser.add_option_group(specific)
	(options, args) = parser.parse_args()
	status = sys.stdout
	orientation = "sense"
	feature = "intervals"
	if options.anti_sense:
		orientation = "anti_sense"
	if options.deletions:
		feature = "deletions"
	if options.substitutions:
		feature = "substitutions"
	output = None
	if options.outputall and options.binoverlap:
		parser.error("cannot combine --outputall and --binoverlap flags\n")
	if options.overlap and options.binoverlap:
		parser.error("cannot combine --overlap and --binoverlap flags\n")
	if options.verbose or not options.output_file:
		status = open("pyBinCollector_log.txt","w")
	if len(sys.argv) < 1:
		parser.error("usage: %prog [options] -f filename -g gene_list. Use -h or --help for options\n")
	binoverlapion = []
	if not options.input_file:
		parser.error("you forgot to include a data file\n")
	if options.output_file:
		output = options.output_file
	if options.binoverlap: binoverlapion = [options.binoverlap[0]-1,options.binoverlap[1]]

	### Making GTF2 parser object and parsing GTF annotation file ###
	status.write("Parsing GTF annotation file...\n")
	gtf = GTF2.Parse_GTF()
	gtf.read_GTF(options.annotation_file,source=options.annotation,transcripts=False)
	###
	
	status.write("Collecting data...\n")

	if options.overlap:
		data = BinCollector()
		data.annotation = options.annotation
		data.findOverlap(options.input_file,gtf,sequence=options.sequence,minfeatlength=options.min_length,maxfeatlength=options.max_length,orientation=orientation,ignorestrand=options.nostrand,out_file=output,ranges=options.range,overlap=options.readoverlap)
	elif options.binoverlap:
		data = BinCollector()
		data.annotation = options.annotation
		data.numberofbins = options.numberofbins
		status.write("Looking for intervals that overlap with bins %s to %s...\n" % (binoverlapion[0]+1,binoverlapion[-1]))
		data.findBinOverlap(options.input_file,gtf,ignorestrand=options.nostrand,sequence=options.sequence,minfeatlength=options.min_length,maxfeatlength=options.max_length,orientation=orientation,out_file=output,printbins=binoverlapion,ranges=options.range,overlap=options.readoverlap)
	elif options.outputall:
		outputfiles = list()
		if options.output_file:
			outfile = open(options.output_file,"w")
		else:
			outfile = open("%s_%s_%s.%s" % (orientation,feature,getfilename(options.input_file),"txt"),"w")
		tempdir = tempfile.mkdtemp(prefix="pyBinCollector_",suffix="_temp",dir="./")
		tempfiles = list()
		status.write("Creating temporary files for each chromosome in the read data...\n")
		datafiles = splitData(options.input_file,tempdir)
		for chromosome in sorted(datafiles):
			status.write("Processing data for chromosome %s...\n" % chromosome)
			temp = tempfile.NamedTemporaryFile(prefix='%s_table' % chromosome,dir=tempdir,suffix='.tmp', delete=False)
			file = datafiles[chromosome]
			table = open(file.name,"r")
			tempfiles.extend([temp,file])
			outputfiles.append(temp.name)
			data = BinCollector()
			data.annotation = options.annotation
			data.numberofbins = options.numberofbins			
			data.countFeatureDensities(table,gtf,sequence=options.sequence,ignorestrand=options.nostrand,minfeatlength=options.min_length,maxfeatlength=options.max_length,feature=feature,orientation=orientation,ranges=options.range,overlap=options.readoverlap,unique=options.unique)
			if data.dataarray:
				if data.numberofbins:
					data.binFeatureDensities(numberofbins=data.numberofbins)
				try:	
					data.printBinCountingResults(out_file=temp.name,normalize=options.normalize)
				except AssertionError:
					status.write("...No data for chromosome %s\n" % chromosome)
			table.close()
		status.write("Merging output files...\n")
		alldata = defaultdict(list)
		for file in outputfiles:
			table = open(file,"r")
			for line in table:
				Fld = line.strip().split()
				gene_name = Fld[0]
				alldata[gene_name] = Fld[1:]
			table.close()
		sortedbylength = sortbylistlength(alldata)
		longestgenelength = sortedbylength[0][1]
		outfile.write("%s\n" % ("\t".join([str(i) for i in range(1,longestgenelength+1)])))
		for gene,length in sortedbylength:
			outfile.write("%s\t%s\n" % (gene,"\t".join([str(i) for i in alldata[gene]])))
		outfile.close()
		status.write("Removing temporary files...\n")
		for i in tempfiles:
			os.unlink(i.name)
		shutil.rmtree(tempdir)		
	else:
		outputfiles = list()
		data = BinCollector()
		data.annotation = options.annotation
		data.numberofbins = options.numberofbins
		if options.output_file:
			outfile = open(options.output_file,"w")
		else:
			outfile = open("%s_%s_%s.%s" % (orientation,feature,getfilename(options.input_file),"txt"),"w")
		status.write("Creating temporary files for each chromosome in the read data...\n")
		tempdir = tempfile.mkdtemp(prefix="pyBinCollector_",suffix="_temp",dir="./")
		datafiles = splitData(options.input_file,tempdir)
		status.write("Calculating cumulative bin densities...\n")	
		tempfiles = list()
		for chromosome in sorted(datafiles):
			data.dataarray = defaultdict(list)
			status.write("Processing data for chromosome %s...\n" % chromosome)
			file  = datafiles[chromosome]
			table = open(file.name,"r")
			tempfiles.append(file)
			data.countFeatureDensities(table,gtf,sequence=options.sequence,ignorestrand=options.nostrand,minfeatlength=options.min_length,maxfeatlength=options.max_length,feature=feature,orientation=orientation,ranges=options.range,overlap=options.readoverlap,unique=options.unique)
			if data.dataarray:
				temp = tempfile.NamedTemporaryFile(prefix='%s_table' % chromosome,dir=tempdir,suffix='.tmp',delete=False)
				tempfiles.append(temp)
				outputfiles.append(temp.name)
				data.binFeatureDensities(numberofbins=data.numberofbins)
				try:
					data.printBinCountingResults(out_file=temp.name,normalize=options.normalize,cumulative=True)
				except AssertionError:
					status.write("...No data for chromosome %s\n" % chromosome)
			table.close()
		cumulativedata = np.zeros(data.numberofbins)
		status.write("Printing results to output file...\n")
		for file in outputfiles:
			cumulativedata += np.loadtxt(file,delimiter="\t",dtype=float)
		outfile.write("# generated by BinCollector, %s\n# %s\n" % (time.ctime(),' '.join(sys.argv)))
		outfile.write("# bin\thits_or_fraction\n")
		for i in range(data.numberofbins):
			outfile.write("%s\t%s\n" % (i+1,cumulativedata[i]))
		outfile.close()
		status.write("Removing temporary files...\n")
		try:
			for i in tempfiles:
				os.unlink(i.name)
			shutil.rmtree(tempdir)
		except:
			status.write("\nERROR! Could not delete the temporary files because of permission issues, please do this manually\n")

	status.write("DONE!\n")
	
if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2018"
__version__		= "0.0.3"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	Clustering.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

from pyCRAC.Parsers.ParseAlignments import ParseCountersOutput
from pyCRAC.Methods import overlap
import numpy as np
import sys

class ReadClustering():
	""" Class for generating read clusters out of read intervals.
	Only accepts GTF files as input file. Made for using with pyReadCounters read interval GTF files. """
	def __init__(self,datafile,logfile=sys.stderr):
		self.data = ParseCountersOutput(datafile,index=True)	# read data file needs to be indexed!
		np.seterrcall(self.__catchErrors)
		np.seterr(all='call')
		self.__end_not_reached = True
		self.__cluster_start = int()
		self.__cluster_end	 = int()
		self.__max_position	 = int()
		self.__max_read_end	 = int()
		self.__line_position = int()
		self.logfile		 = logfile
		self.index			 = None
		self.coveragearray	 = None
		self.subsarray		 = None
		self.delsarray		 = None
		self.chromosome		 = str()
		self.strand			 = str()
		self.read_length	 = 0
		self.line_count		 = 0
		self.cluster_height	 = 0
		self.cluster_length	 = 0
		self.cluster_count	 = 0
		self.cluster_read_count = 0
		self.max_cluster_length = int()
	
	def __catchErrors(self,type,flag):
		""" catches any numpy floating point errors and prints the problem cases to the standard error"""
		self.logfile.write("\nZero division error (%s), with flag %s\n%s" % (type,flag,self.data.line))
		self.logfile.write("subsarray:\n%s\ndelsarray:\n%s\ncoverageaarray:\n%s\n" % (self.subsarray,self.delsarray,self.coveragearray))
		
	def __trimNumpyArrays(self):
		""" shrinks the numpy arrays by removing any zeros """
		position = np.where(self.coveragearray > 0)[0][-1] + 1
		self.coveragearray = self.coveragearray[:position]
		self.subsarray	   = self.subsarray[:position]
		self.delsarray	   = self.delsarray[:position]
		
	def mutsAnnotationString(self,min_frequency=0):
		"""Makes a CIGAR-type string for mutations in reads or clusters"""
		subsstring = str()
		delsstring = str()
		mutlist	   = list()
		subsfreqarray,delsfreqarray = self.mutsCountsToFrequencies(min_frequency=float(min_frequency))
		for nr in range(len(subsfreqarray)):
			if subsfreqarray[nr]:
				subsstring = "S%.1f" % subsfreqarray[nr]
			if delsfreqarray[nr]:
				delsstring = "D%.1f" % delsfreqarray[nr]
			mutsstring = "%s%s%s" % (self.index[nr]+1,subsstring,delsstring) 
			if subsstring or delsstring:
				mutlist.append(mutsstring)
				subsstring = str()
				delsstring = str()
		if mutlist:
			return ','.join(mutlist)
		else:
			return None

	def mutsCountsToFrequencies(self,min_frequency=0):
		""" uses the cluster data to convert mutation numbers to mutation frequencies for each position """
		min_frequency = float(min_frequency)	# set the minimal mutation frequency
		subsfreqarray = self.subsarray * 100 / self.coveragearray # make an array of substitution frequencies
		delsfreqarray = self.delsarray * 100 / self.coveragearray # make an array of deletion frequencies
		substitutions = np.where(subsfreqarray > min_frequency,subsfreqarray,0)		# return all the positions larger than the min frequency
		deletions	  = np.where(delsfreqarray > min_frequency,delsfreqarray,0)
		return substitutions,deletions

	def __startNewCluster(self):
		""" method for setting up a new cluster. Note that the method assumes that you have already read a new line from the data file."""
		self.__start_new_cluster = False					# indicate that no new cluster has to be made in the next iteration.
		self.subsarray	= np.zeros(self.max_cluster_length)
		self.delsarray	= np.zeros(self.max_cluster_length)
		self.coveragearray = np.zeros(self.max_cluster_length)
		self.index		= np.arange(self.__cluster_start,self.__cluster_start+self.max_cluster_length)
		self.chromosome = self.data.chromosome
		self.strand		= self.data.strand
		self.start		= int()
		self.end		= int()
		self.cluster_height = 1
		self.cluster_read_count = 1
		if not self.__cluster_start:
			self.__cluster_start = self.data.read_start
		self.__max_position = self.__cluster_start + self.max_cluster_length
		self.__cluster_end = self.data.read_end
		self.__updateCluster()	# update the arrays with the information from the first read in the cluster. 
 
	def __finishCluster(self):
		""" When a cluster has been made it corrects the length of the arrays. It also keeps track of the number of clusters made. """
		self.cluster_length = self.end - self.start						# get the cluster length.
		self.subsarray = self.subsarray[:self.cluster_length]			# update the length of the arrays to make them as long as the cluster.
		self.delsarray = self.delsarray[:self.cluster_length]
		self.coveragearray = self.coveragearray[:self.cluster_length]
		self.cluster_count += 1											# update the cluster count.
		
	def __getNewRead(self):
		self.__end_not_reached = self.data.readLineByLine(numpy=True)	# get a line from the read data file
		self.read_length = self.data.read_end - self.data.read_start	# determine the read length
		self.line_count +=1												# add 1 to the read counter.
			
	def __updateCluster(self):
		""" adds the read information to the arrays."""
		self.data.number_of_reads
		self.coveragearray += np.where(((self.index <= self.data.read_end) & (self.index >= self.data.read_start)),self.data.number_of_reads,0)
		self.subsarray[np.in1d(self.index,self.data.substitutions)] += self.data.number_of_reads	# update the substitutions array.
		self.delsarray[np.in1d(self.index,self.data.deletions)] += self.data.number_of_reads		# update the deletions array.
		self.cluster_read_count += self.data.number_of_reads						                # update the cluster read count.
	
	def getCluster(self,min_overlap=1,min_cdnas=2,max_length=100,min_height=2,min_length=10,debug=False):
		"""Outputs strings of sequences that contain at least two single overlapping reads and it ignores repeated sequences. Behaves as an iterator.
		NOTE! For the clustering to work properly the GTF file has to be sorted in a specific way! First by chromosome, then by strand and then by start
		position! Use the sort -k1,1 -k7 -k4,4n to sort your GTF file. Returns 'True' if it found a new cluster and has updated all the class variables. 
		For each cluster it generates numpy arrays for substitutions deletions and the coverage (self.subsarray,self.delsarray and self.coveragearray). 
		It also keeps track of the read count whilst reading through the read gtf file. 
		You can put this method in a while loop as it returns 'False' when it reaches the end of the file.
		while object.getCluster():
			do something.....
				cluster_start = object.start
				cluster_end = object.end
				coverage  = object.coveragearray
				etc...
		Requires the file to be sorted first by chromosome and then by start position!!!!
		The maximum cluster length is set at 100 bp but this can be changed by setting the max_length variable."""
		self.max_cluster_length = max_length
		while self.__end_not_reached:
			if not self.line_count:									# start a new cluster if the first line in the file has not been read yet.
				self.__getNewRead()									# get a line from the file.
				self.__startNewCluster()							# start a new cluster.
			elif self.__start_new_cluster:							# if the cluster has been returned, start a new one.
				self.__startNewCluster()							# start a new cluster if a cluster has been returned.
			else:													# if a start to a new cluster has been made, just get another read.
				self.__getNewRead()									# get a line from the file.
				if self.data.chromosome == self.chromosome and self.data.strand == self.strand and overlap(self.data.read_start,self.data.read_end-1,self.__cluster_start,self.__cluster_end,overlap=min_overlap):	# if they are on the same chromosome and strand and overlap, continue
					self.__updateCluster()									# update the arrays with the read information.
					self.__start_new_cluster = False						# we don't have to start thinking about a new cluster as long as there is overlap.
					if self.data.read_end > self.__max_position:			# if the end position of the last read is beyond the cluster boundary, note position of the read in the file and continue.
						if not self.__line_position:
							self.__line_position = self.line_count - 1		# it needs to remember on which line this occurred. After finishing the cluster, it needs to go back to these lines for the next cluster.
						self.__cluster_end = self.__max_position			# but the end position of the cluster can already be set here.
					else:
						if self.data.read_end >= self.__cluster_end:		# if the read falls completely within the cluster boundaries, just continue.
							self.__cluster_end = self.data.read_end			# if not, just set the end position of the cluster.
				else:
					self.start = self.__cluster_start						# set the start position of the current cluster.
					self.end   = self.__cluster_end							# set the end position of the current cluster.
					self.__start_new_cluster = True							# indicate that a new cluster has to be made in the next iteration.
					if self.__line_position:								# if there is no longer overlap,
						self.data.seek(self.__line_position)				# go back to position where the end position was larger than the cluster boundary and start from there for the next cluster.
						self.line_count = self.__line_position				# after seek, the read count needs to be set back to that position.
						self.__getNewRead()									# then re-read the interval data at that position.
						self.__line_position = int()						# reset the line position variable.
						self.__cluster_start = self.__max_position + 1		# set the new cluster start.
					else:
						self.__cluster_start = self.data.read_start			# set the start position of the next cluster
					#### does the existing cluster fit all the criteria? ###	  
					self.cluster_height = int(self.coveragearray.max())										# calculate the cluster height.
					if self.cluster_read_count >= min_cdnas and self.cluster_height >= min_height:			# if the cluster fits all the criteria, return True.
						self.__finishCluster()																# adjust the length of the arrays and finish the cluster.
						if self.cluster_length >= min_length:
							return True																		# indicate that a cluster was found.
						


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.2.3"
__credits__		= ["Sander Granneman", "Shaun Webb"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyReadAligner.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
from optparse import *
from pyCRAC.Parsers import GTF2
from pyCRAC.Classes.Aligner import *

def main():
	parser = OptionParser(usage="usage: %prog [options] -f filename -g gene_list", version="%s" % __version__)
	files = OptionGroup(parser,"File input options")
	files.add_option("-f", "--input_file", dest="filename",help="As input files you can use Novoalign native output or SAM files as input file. By default it expects data from the standard input. Make sure to specify the file type of the file you want to have analyzed using the --file_type option!", metavar="FILE",default=None)
	files.add_option("-o", "--output_file", dest="output_file",help="Use this flag to override the standard output file names. All alignments will be written to one output file.",default=None)
	files.add_option("-g", "--genes_file", dest="genes",help="here you need to type in the name of your gene list file (1 column) or the hittable file", metavar="FILE",default=None)
	files.add_option("--chr", dest="chr_coord",help="if you simply would like to align reads against a genomic sequence you should generate a tab delimited file containing an identifyer, chromosome name, start position, end position and strand ('-' or '+')", metavar="FILE",default=None)
	files.add_option("--gtf",dest="annotation_file",metavar="annotation_file.gtf",help="type the path to the gtf annotation file that you want to use",default=None)
	files.add_option("--tab",dest="tab_file",metavar="tab_file.tab",help="type the path to the tab file that contains the genomic reference sequence",default=None)
	files.add_option("--file_type",dest="file_type",choices=["novo","sam","gtf"],help="use this option to specify the file type (i.e. 'novo', 'sam', 'gtf'). This will tell the program which parsers to use for processing the files. Default = 'novo'",default="novo")
	specific = OptionGroup(parser,"pyReadAligner specific options")
	specific.add_option("--limit",dest="limit",metavar="500",type="int",help="with this option you can select how many reads mapped to a particular gene/ORF/region you want to count. Default = All",default=-1)
	common = OptionGroup(parser, "Common options")
	common.add_option("-v","--verbose",dest="verbose",action="store_true",help="prints all the status messages to a file rather than the standard output",default=False)
	common.add_option("--ignorestrand",dest="nostrand",action="store_true",help="this flag tells the program to ignore strand information and all overlapping reads will considered sense reads. Useful for analysing ChIP or RIP data",default=False)
	common.add_option("--zip", dest="zip", metavar="FILE", help="use this option to compress all the output files in a single zip file", default=None)
	common.add_option("--overlap",dest="overlap",help="sets the number of nucleotides a read has to overlap with a gene before it is considered a hit. Default =  1 nucleotide",type="int",metavar="1",default="1")
	common.add_option("-s","--sequence",dest="sequence",metavar="genomic",help="with this option you can select whether you want the reads aligned to the genomic or the coding sequence. Default = genomic",default="genomic")
	common.add_option("-r","--range",dest="range",type="int",metavar="100",help="allows you to set the length of the UTR regions. If you set '-r 50' or '--range=50', then the program will set a fixed length (50 bp) regardless of whether the GTF file has genes with annotated UTRs.",default=False)
	novo = OptionGroup(parser, "Options for novo, SAM and BAM files")
	novo.add_option("--align_quality", "--mapping_quality", dest="align_quality", metavar="100", type="int", help="with these options you can set the alignment quality (Novoalign) or mapping quality (SAM) threshold. Reads with qualities lower than the threshold will be ignored. Default = 0", default=0)
	novo.add_option("--align_score", dest="align_score", metavar="100", type="int", help="with this option you can set the alignment score threshold. Reads with alignment scores lower than the threshold will be ignored. Default = 0", default=0)
	novo.add_option("-l","--length",dest="length",metavar="100",type="int",help="to set read length threshold. Default = 1000",default=1000)
	novo.add_option("-m","--max",dest="max_reads",metavar="100000",type="int",help="maximum number of mapped reads that will be analyzed. Default = All",default=float("infinity"))
	novo.add_option("--unique",dest="unique",action="store_true",help="with this option reads with multiple alignment locations will be removed. Default = Off",default=False)		
	novo.add_option("--blocks",dest="blocks",action="store_true",help="with this option reads with the same start and end coordinates on a chromosome will only be counted once. Default = Off",default=False)
	novo.add_option("--discarded",dest="discarded",metavar="FILE", help="prints the lines from the alignments file that were discarded by the parsers. This file contains reads that were unmapped (NM), of poor quality (i.e. QC) or paired reads that were mapped to different chromosomal locations or were too far apart on the same chromosome. Useful for debugging purposes",default=None)
	novo.add_option("-d","--distance",dest="distance",metavar="1000",type="int",help="this option allows you to set the maximum number of base-pairs allowed between two non-overlapping paired reads. Default = 1000",default=1000)
	novo.add_option("--mutations",dest="muts",metavar="delsonly",help="Use this option to only track mutations that are of interest. For CRAC data this is usually deletions (--mutations=delsonly). For PAR-CLIP data this is usually T-C mutations (--mutations=TC). Other options are: do not report any mutations: --mutations=nomuts. Only report specific base mutations, for example only in T's, C's and G's :--mutations=[TCG]. The brackets are essential. Other nucleotide combinations are also possible",default=None)
	parser.add_option_group(files)
	parser.add_option_group(specific)
	parser.add_option_group(common)
	parser.add_option_group(novo)
	(options, args) = parser.parse_args()
	status = sys.stdout
	if options.verbose:
		status = open("pyReadAligner_log.txt","w")
	if not options.filename:
		options.filename = sys.stdin
	if options.muts:
		mutslist = re.compile("\[([A-Za-z]+)\]")
		mutstype = mutslist.findall(options.muts)
		if mutstype: options.muts = [i for i in mutstype[0]]
	if len(sys.argv) < 2:
		parser.error("usage: %prog [options] -f filename -g genes_list. Use -h or --help for options\n")
	if options.genes is None and options.chr_coord is None:
		parser.error("you forgot to input your genes list, hittable file or chromosome coordinate file. Use -g, --genes or --chr options\n")
	if options.tab_file is None:	
		parser.error("you forgot to input the tab file containing the genomic references sequence. Use --tab=yourfavoritetab.tab option\n")
	if options.filename:
		status.write("processing %s data...\n" % options.file_type)
	
	genes = list()
	
	### Making GTF2 parser object and parsing GTF annotation file ###
	gtf = GTF2.Parse_GTF()
	###
	
	### processing list of genes or chromosome coordinate files:
	if options.genes:
		status.write("parsing GTF reference file...\n")
		genes = [line.strip() for line in open(options.genes,"r").readlines()]
		gtf.read_GTF(options.annotation_file,ranges=options.range,genes=genes,transcripts=True)
	
	elif options.chr_coord:
		status.write("processing chromosomal coordinates file...\n")
		for coord in open(options.chr_coord,"r").readlines():
			gene,chromosome,start,end,strand =	coord.strip().split()
			start = int(start)
			end	  = int(end)
			genes.append(gene)
			gtf.addGene(gene,start,end,"exon",None,chromosome,strand)
	###
	
	### storing chromosomes into memory:
	chromosomes = [gtf.chromosome(gene) for gene in genes]
	status.write("storing relevant reference sequences into memory...\n")
	gtf.read_TAB(options.tab_file,chromosomes)	# only the chromosomes that are required are stored into memory.
	###
		 
	### creating a ReadAligner object:	 
	solexadata = ReadAligner(options.filename,gtf,genes,file_type=options.file_type,sequence=options.sequence,blocks=options.blocks,debug=options.discarded)
	###
	
	### Aligning the reads:
	solexadata.alignReads(length=options.length,align_quality=options.align_quality,align_score=options.align_score,MAX=options.max_reads,limit=options.limit,unique=options.unique,muts_filter=options.muts)
	###
	
	### if no sense or anti-sense reads were mapped to any of our genes, report this to the log or standard output
	nosensehits		= list()
	noantisensehits = list()
	for gene in genes:
		if gene not in solexadata.sense_seqs:
			nosensehits.append(gene)
		if gene not in solexadata.anti_sense_seqs:
			noantisensehits.append(gene)
	if nosensehits	  : status.write("Could not find any sequences overlapping 'sense' with genes %s...\n" % (','.join(nosensehits)))
	if noantisensehits: status.write("Could not find any sequences overlapping 'anti-sense' with genes %s...\n" % (','.join(noantisensehits)))
	###
	
	status.write("printing output files...\n")
	solexadata.printFastaWithAlignments(out_file=options.output_file)
	
	### compressing output files, if requested.
	if options.zip:					
		import zipfile
		status.write("compressing files...\n")
		file = zipfile.ZipFile(options.zip,mode="w",compression=zipfile.ZIP_DEFLATED)
		for files in solexadata.file_list:
			file.write(files)
		file.close()	
		for files in solexadata.file_list:
			try:
				os.remove(files)
			except OSError:
				pass	
	###
		
if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__		= 'Sander Granneman'
__copyright__	= 'Copyright 2019'
__version__		= '0.0.6'
__credits__		= ['Sander Granneman']
__maintainer__	= 'Sander Granneman'
__email__		= 'sgrannem@ed.ac.uk'
__status__		= 'Production'

##################################################################################
#
#	pyFastqJoiner.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the 'Software'), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

"""
Joins two fastq files into one. Useful when you need to flatten or collaps paired end data

"""

import sys
import six
import gzip
from optparse import *
from pyCRAC.Methods import reverse_complement

def mergeFastqFiles(forward,reverse,filetype,merged=None,character=None,compressed=False,reversecomplement=False):
	''' merges the sequence and quality values of the two fastq files and returns a new fastq file. 
	This tool does NOT check whether the header names are the same, it simply merges them.
	When setting reverse_complement=True, the reverse reads will be reverse complemented before merging.'''
	
	### processing the input files ###

	fastq_f = None
	fastq_r = None

	supportedfiletypes = ['fastq','fastq.gz']
	
	if filetype == supportedfiletypes[0]:	   # if the input files are not compressed
		fastq_f	 = open(forward,'r')
		fastq_r	 = open(reverse,'r')
	elif filetype == supportedfiletypes[1]:	   # if the input files were compressed using gunzip
		fastq_f	 = gzip.open(forward,'rb')
		fastq_r	 = gzip.open(reverse,'rb')
	else:
		raise RuntimeError('\n\nUnrecognizable file type %s\n' % filetype)
	
	### preparing the output file. It is also possible to pipe the output to another script by not naming the output file ###
	
	if merged and not compressed:
		output	= open(merged,'w')
	elif merged and compressed:										   
		output	= gzip.open('%s.gz' % merged,'wb')
	else:
		output	= sys.stdout
	
	### iterating over the input files and merging the records ###		   
	
	files = list()
	files.extend([fastq_f,fastq_r,output])
	while True:
		try:
			line = str()
			name_f, name_r = str(six.next(fastq_f).strip()), str(six.next(fastq_r).strip())
			seq_f, seq_r   = str(six.next(fastq_f).strip()), str(six.next(fastq_r).strip())
			six.next(fastq_f)
			six.next(fastq_r)
			qual_f, qual_r = str(six.next(fastq_f).strip()), str(six.next(fastq_r).strip())

			if reversecomplement:
				seq_r  = reverse_complement(seq_r)	# reverse-complement the sequence string
				qual_r = qual_r[::-1]				# reversing the quality information				

			line = '%s%s\n%s%s%s\n+\n%s%s\n' % (name_f,name_r,seq_f,character,seq_r,qual_f,qual_r)

			if compressed:
				line = line.encode()
			output.write(line)

		except StopIteration:
			break
	for i in files:
		try:
			i.close()  
		except ValueError:
			pass   
		
def mergeFastaFiles(forward,reverse,filetype,merged=None,character=None,compressed=False,reversecomplement=False):
	''' merges the sequence and quality values of the two fasta files and returns a new fasta file. 
	NOTE!!!!This tool does NOT check whether the header names are the same, it simply merges them,'''
	
	### processing the input files ###

	fasta_f = None
	fasta_r = None
	supportedfiletypes = ['fasta','fasta.gz']
	
	if filetype == supportedfiletypes[0]:	   # if the input files are not compressed
		fasta_f	 = open(forward,'r')
		fasta_r	 = open(reverse,'r')
	elif filetype == supportedfiletypes[1]:	   # if the input files were compressed using gunzip
		fasta_f	 = gzip.open(forward,'rb')
		fasta_r	 = gzip.open(reverse,'rb')
	else:
		raise RuntimeError('\n\nUnrecognizable file type %s\n' % filetype)
	
	### preparing the output file. It is also possible to pipe the output to another script by not naming the output file ###		 
	
	if merged and not compressed:
		output	= open(merged,'w')
	elif merged and compressed:										   
		output	= gzip.open('%s.gz' % merged,'wb')
	else:
		output	= sys.stdout
	
	### iterating over the input files and merging the records ###		   
	
	files = list()
	files.extend([fasta_f,fasta_r,output])
	while True:
		try:
			line = str()
			name_f, name_r	= str(six.next(fasta_f).strip()),str(six.next(fasta_r).strip())
			seq_f,	seq_r	= str(six.next(fasta_f).strip()),str(six.next(fasta_r).strip())
			if reversecomplement:
				seq_r = reverse_complement(seq_r)	# reverse-complement the sequence string

			line = '%s%s\n%s%s%s\n' % (name_f,name_r,seq_f,character,seq_r)

			if compressed:
				line = line.encode()

			output.write(line)
				
		except StopIteration:
			break

	for i in files:
		try:
			i.close()  
		except ValueError:
			pass   
def main():
	parser = OptionParser(usage='usage: %prog [options] -f file1 file 2', version="%s" % __version__)
	parser.add_option('-f', '--filename',nargs=2,metavar='fastq_file1 fastq_file2',dest='filename',help='provide the names of two raw data files separated by a single space')
	parser.add_option('--file_type',choices=['fasta','fastq','fasta.gz','fastq.gz'],dest='filetype',help='type of file, uncompressed or compressed (fasta, fasta.gz, fastq.gz, (gzip/gunzip compressed)) fastq. Default is fastq', metavar='FASTQ',default='fastq')
	parser.add_option('--reversecomplement',dest='reversecomplement',action='store_true',help='to reverse-complement the reverse read before joining. Default is False',default=False)
	parser.add_option('-o', '--outfile',metavar='mergedfastq.fastq',dest='outfile',help='provide the name of the output file. By default it will be printed to the standard output',default=None)
	parser.add_option('-c', '--character',metavar="|",dest='character',help='This option adds the "|" character between the DNA sequences so that it is much easier to split the data again later on',default="")
	parser.add_option('--gz','--gzip', dest='gzip', action='store_true',help='use this option to compress all output file using gunzip or gzip (compression level 9). No need to add the .gz extension. Note that this slows down the program significantly', default=False)
	(options, args) = parser.parse_args()
	if not options.filename or len(options.filename) < 2:
		parser.error('you need to input two filenames!\nExample: -f file1.fastq file2.fastq\n')
	supportedfiletypes = ['fasta','fasta.gz','fastq','fastq.gz']
	if options.filetype in supportedfiletypes[:2]:
		mergeFastaFiles(options.filename[0],options.filename[1],options.filetype,merged=options.outfile,character=options.character,compressed=options.gzip,reversecomplement=options.reversecomplement)
	else:
		mergeFastqFiles(options.filename[0],options.filename[1],options.filetype,merged=options.outfile,character=options.character,compressed=options.gzip,reversecomplement=options.reversecomplement)
	
if __name__ == '__main__':
	main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2018"
__version__		= "1.2.1"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyAlignment2Tab.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
from optparse import *
from collections import defaultdict
from pyCRAC.Methods import getfilename

def color(string):
	"""GREY, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
	NONE, BOLD, FAINT, ITALIC, UNDERLINE, BLINK, FAST, REVERSE, CONCEALED = range(8)
	#The background is set with 40 plus the number of the color, and the foreground with 30
	#These are the sequences need to get colored ouput
	RESET_SEQ = "\033[0m"
	COLOR_SEQ = "\033[1;%dm"
	BOLD_SEQ = "\033[1m"
	"""
	if re.search("H",string):
		return string
	else:
		a = list(string)
		for nr,i in enumerate(a):
			if i.islower(): # bold
				a[nr] = '\033[1m%s\033[0m' % i
			if i == "A":	# bold letters on a red background
				a[nr] = '\033[1;41;38m%s\033[0m' % i
			if i == "C":	# bold letters on a green background
				a[nr] = '\033[1;42m%s\033[0m' % i
			if i == "G":	# bold letters on a yellow background
				a[nr] = '\033[1;43m%s\033[0m' % i
			if i == "T":	# bold lettters on a green background
				a[nr] = '\033[1;44m%s\033[0m' % i
		return ''.join(a)

def countHeadGaps(string):
	""" counts the number of '-' in the beginning of the string """
	a = re.search(r'[atgcnATGCN]',string)
	return a.start()

def sortByHeadGaps(dictionary):
	""" returns a list of sorted sequences """
	return sorted([(countHeadGaps(dictionary[keys]),keys,dictionary[keys]) for keys in dictionary])
	
def printSeqGaps(sequence):
	""" prints sequence gap characters """
	string = str()
	for i in range(1,len(sequence)+1):
		if i %10 == 0:
			string += "|"
		else:
			string += "."
	return string

def printTab(filein,fileout,limit=90,removeempty=False):
	""" prints the output in the tabular format """
	fasta_dict = dict()
	readid	   = str()
	for line in filein:
		if line.startswith(">"):
			readid = line.strip()
		else:
			fasta_dict[readid] = line.strip()
	sorted_tuple_list = sortByHeadGaps(fasta_dict)
	headerlength = int(max([len(y) for x,y,z in sorted_tuple_list]))
	seq_dots   = printSeqGaps(sorted_tuple_list[0][2])
	seq_length = int(len(sorted_tuple_list[0][2]))
	frag_nr	   = int(seq_length/limit)
	for i in range(frag_nr+1):
		end = int()
		if i == frag_nr:
			end = seq_length
		else:
			end = ((i+1)*limit)
		position = str((i*limit)+1)
		adjustment = headerlength - len(position)
		position = ' ' * adjustment + position
		fragmentlist = list()
		for (nr,header,sequence) in sorted_tuple_list:
			fragment = sequence[i*limit:(i+1)*limit:]
			if re.search("[atcgATCG.]",fragment):
				adjustment = headerlength - len(header)
				header = ' ' * adjustment + header
				if fileout is sys.stdout:
					fragmentlist.append("%s\t%s\n" % (header,color(fragment)))
				else:
					fragmentlist.append("%s\t%s\n" % (header,fragment))	
		if len(fragmentlist) <= 1 and removeempty:
			continue
		fileout.write("%s\t%s %d\n" % (position,seq_dots[i*limit:(i+1)*limit:],end))			
		fileout.write("%s\n" % "".join(fragmentlist))
					
def parseSingleFile(filein):
	""" for the Galaxy server. In case all the alignments were printed to a single file. """
	datadict = defaultdict(str)
	for line in filein:
		if not line.strip():	# if the parser encounters an empty line, continue
			continue
		if line.startswith("###"):
			gene = line.strip()
			datadict[gene] = list()
		else:
			datadict[gene].append(line)
	return datadict
			
					
def main():
	parser = OptionParser(usage="usage: %prog -f myfastafile -o mytabfile --limit=90", version="%s" % __version__)
	parser.add_option("-f",dest="infile",metavar="data.fasta",help="type the path to the fasta file that you want to use. By default it expects data from the standard input.",default=None)
	parser.add_option("-o","--outfile",dest="outfile",metavar="data.tab",help="type the name and path of the file you want to write the output to. Default is standard output",default=None)
	parser.add_option("--limit",dest="limit",metavar="90",type="int",help="Allows the user to set the column width of the alignment. Default=90 characters",default=90)
	parser.add_option("--noempty",dest="noempty",action="store_true",help="only print reference sequences that have reads mapped to it. Default is False",default=False)
	parser.add_option("--singlefile",dest="singlefile",action="store_true",help="include this flag if all your alignments are in a single file, separated by empty lines",default=False)	
	(options, args) = parser.parse_args()
	data = sys.stdin
	out	 = sys.stdout
	if options.outfile:
		out = open(options.outfile,"w")
	if options.infile:
		data = open(options.infile,"r")
	if options.singlefile:
		datadict = parseSingleFile(data)
		for gene in datadict:
			out.write("\n%s\n" % gene)
			printTab(datadict[gene],out,limit=options.limit,removeempty=options.noempty)
	else:
		printTab(data,out,limit=options.limit,removeempty=options.noempty)

if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "1.4.4"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "production"

##################################################################################
#
#	Counters.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import time
from pyCRAC.Parsers.ParseAlignments import *
from pyCRAC.Methods import strand_converter, overlap, numpy_overlap, getfilename
from pyCRAC.Classes.Exceptions import *
from pyCRAC.Classes.NGSFormatWriters import NGSFileWriter

class PyCounters(ParseAlignmentFile,ParseCountersOutput):
	def __init__(self,file_in,gtfobject,file_type="novo",ranges=0,ignorestrand=False,debug=None):
		"""Initializing the class attributes"""
		file_types = ["novo","sam","gtf"]
		if file_type in file_types[:2]:
			ParseAlignmentFile.__init__(self,file_in,file_type,debug=debug)
		elif file_type == file_types[2]:
			ParseCountersOutput.__init__(self,file_in)
		self.gtf				   = gtfobject
		self.file_name			   = getfilename(file_in)
		self.datadictionary		   = defaultdict(lambda: defaultdict(lambda: [0,0]))	# annotation is first key, then gene, then the values is sense and anti_sense counts
		self.gene_hits			   = defaultdict(int)
		self.cDNA_count			   = defaultdict(int)
		self.annotations_count	   = defaultdict(int)
		self.hits_count			   = int()
		self.no_matches			   = int()
		self.sense_hits_count	   = int()
		self.anti_sense_hits_count = int()
		self.file_list			   = list()
		self.sequence              = "genomic"
		self.densities             = False
		self.__ignorestrand		   = ignorestrand

	def countGeneHits(self,min_overlap=1,sense=True,anti_sense=True,sequence="genomic"):
		"""Calculates the number of hits for genes in the dataset. It does NOT count transcripts! To map cDNAs/clusters to individual
		transcripts you can run the countGeneFeaturesOverlap after running the countGeneHits method. The program first determines to which
		genes hits were found and then with the countGeneFeaturesOverlap you can calculate overlap for individual gene transcripts and
		transcript features such as intron,UTR, exon and CDS. This saves a tremendous number of iterations.
		The GTF2 module calculates start and end positions for each gene based exon coordinate and include introns and UTRs if present.
		This makes the iteration step significantly faster, as in higher eukaryotes each gene can code for many transcripts"""
		list_of_tuples = defaultdict(list)
		self.sequence = sequence
		for interval in self.count:
			selectedgenes = list()
			chromosome,strand,read_start,read_end,substitutions,deletions = interval
			coverage = self.count[interval]["coverage"]
			strand = strand_converter(strand)
			if not self.count[interval]["features"]:
				search_results = list()
				try:
					if chromosome not in list_of_tuples:
						list_of_tuples[chromosome] = self.gtf.chromosomeGeneCoordIterator(chromosome,numpy=True,sequence=sequence)
					if len(list_of_tuples[chromosome]) > 0:
						search_results = numpy_overlap(list_of_tuples[chromosome],read_start,read_end-1,overlap=min_overlap)
					else:
						continue
				except AssertionError:			# in case you want to run pyReadCounters with a smaller GTF. This should ignore any errors
					#sys.stderr.write("AssertionError @ %s\t%s\t%s\t%s\n" % (chromosome,strand,read_start,read_end))
					pass
				if search_results:
					self.count[interval]["features"] = search_results
			if self.count[interval]["features"]:
				for gene in self.count[interval]["features"]:
					try:
						if sense and strand == self.gtf.strand(gene):
							for i in self.gtf.annotations(gene):
								self.annotations_count[i] += coverage
								self.datadictionary[i][gene][0] += coverage
							self.gene_hits[gene] += coverage
							self.hits_count += coverage
							self.sense_hits_count += coverage
							self.cDNA_count[gene] += 1
							selectedgenes.append(gene)
						if anti_sense and strand != self.gtf.strand(gene):
							for i in self.gtf.annotations(gene):
								self.annotations_count[i] += coverage
								self.datadictionary[i][gene][1] += coverage
							self.gene_hits[gene] += coverage
							self.hits_count += coverage
							self.anti_sense_hits_count += coverage
							self.cDNA_count[gene] += 1
							selectedgenes.append(gene)
					except LookupError:
						#sys.stderr.write('could not finde gene %s in the gtf file\n')
						continue
			self.count[interval]["features"] = selectedgenes

	def countNucleotideDensities(self,min_overlap=1,sense=True,anti_sense=True,sequence="genomic"):
			"""Calculates the number of nucleotides that overlap with features in genes """
			list_of_tuples = defaultdict(list)
			self.sequence = sequence
			self.densities = True
			for interval in self.count:
				selectedgenes = list()
				chromosome,strand,read_start,read_end,substitutions,deletions = interval
				coverage = self.count[interval]["coverage"]
				strand = strand_converter(strand)
				search_results = list()
				try:
					if chromosome not in list_of_tuples:
						list_of_tuples[chromosome] = self.gtf.chromosomeGeneCoordIterator(chromosome,numpy=True,sequence=sequence)
					if len(list_of_tuples[chromosome]) > 0:
						search_results = numpy_overlap(list_of_tuples[chromosome],read_start,read_end-1,overlap=min_overlap,returnall=True)
					else:
						continue
				except AssertionError:			# in case you want to run pyReadCounters with a smaller GTF. This should ignore any errors
					#sys.stderr.write("AssertionError @ %s\t%s\t%s\t%s\n" % (chromosome,strand,read_start,read_end))
					pass
				if search_results:
					for result in search_results:
						gene_start,gene_end,gene = result
						#print "gene start: %s\tgene end: %s\t gene: %s" % (gene_start,gene_end,gene)
						coordinates = np.arange(read_start,read_end)
						#print "read_start: %s\tread_end: %s" % (read_start,read_end)
						#print "coordinate length = %s" % len(coordinates)
						nucleotides = len(coordinates[(coordinates <= gene_end) & (coordinates >= gene_start)])
						#print "overlap = %s" % nucleotides
						try:
							if sense and strand == self.gtf.strand(gene):
								for i in self.gtf.annotations(gene):
									self.annotations_count[i] += nucleotides
									self.datadictionary[i][gene][0] += nucleotides
								self.gene_hits[gene] += nucleotides
								self.hits_count += coverage
								self.sense_hits_count += nucleotides
								self.cDNA_count[gene] += 1
								selectedgenes.append(gene)
							if anti_sense and strand != self.gtf.strand(gene):
								for i in self.gtf.annotations(gene):
									self.annotations_count[i] += nucleotides
									self.datadictionary[i][gene][1] += nucleotides
								self.gene_hits[gene] += nucleotides
								self.hits_count += coverage
								self.anti_sense_hits_count += nucleotides
								self.cDNA_count[gene] += 1
								selectedgenes.append(gene)
						except LookupError:
							continue
				self.count[interval]["features"] = selectedgenes


	def printHitTable(self,out_file=None,rpkm=False):
		"""Does what it says. You have to run at least countReads and countGeneHits methods first before you can use the printHitTable method. You can manually set
		the name of the output file using the 'out_file' argument. Note that once you have used this method the dictionary that stores all the gene information will be destroyed.
		This is done to preserve memory."""
		assert self.hits_count, "\n\nYou need to run the countGeneHits method first before you can use the printHitTable method.\n"
		self.__setInsertVariables()
		file_type = "hittable"
		if out_file:
			self.file_name = out_file
		file_name = self.__createFileHandles(file_type)
		file_out  = open(file_name, "w")
		file_out.write("# generated by Counters version %s, %s\n" % (__version__,time.ctime()))
		file_out.write("# %s\n" %(' '.join(sys.argv)))
		file_out.write("# total number of reads\t%d\n" % self.total_reads)
		file_out.write("# total number of mapped %s:\t%d\n" % (self.__insert,self.__insert2))
		file_out.write("# total number of paired reads\t%d\n" % self.paired_reads)
		file_out.write("# total number of single reads\t%d\n" % self.single_reads)
		file_out.write("# total number of %s overlapping genomic features\t%d\n" % (self.__insert3,self.hits_count))
		file_out.write("# \tsense\t%d\n" % (self.sense_hits_count))
		file_out.write("# \tanti-sense\t%d\n" % (self.anti_sense_hits_count))
		if rpkm:
			file_out.write("# feature\tsense_overlap\tsense-RPKM\tanti-sense_overlap\tanti-sense-RPKM\tcDNA count\n")
		else:
			file_out.write("# feature\tsense_overlap\tanti-sense_overlap\tcDNA count\n")		 
		for annotation,count in sorted(list(self.annotations_count.items()), key=lambda k_v: k_v[1],reverse=True):
			total_sense,total_anti_sense = [sum(i) for i in zip(*iter(self.datadictionary[annotation].values()))]
			if rpkm:
				total_sense_rpkm = int()
				total_anti_sense_rpkm = int()
				for gene,(sense,anti_sense) in sorted(self.datadictionary[annotation].items()):
					total_sense_rpkm += self.__calcRPKM(gene,sense)
					total_anti_sense_rpkm += self.__calcRPKM(gene,anti_sense)
				file_out.write("\n## %s\t%s\t%.3f\t%s\t%.3f\n" % (annotation,total_sense,total_sense_rpkm,total_anti_sense,total_anti_sense_rpkm))
			else:
				file_out.write("\n## %s\t%s\t%s\n" % (annotation,total_sense,total_anti_sense))
			for gene,(sense,anti_sense) in sorted(list(self.datadictionary[annotation].items()), key=lambda k_v1_v2: k_v1_v2[1][0],reverse=True):
				if rpkm:
					file_out.write("%s\t%s\t%.3f\t%s\t%.3f\t%s\n" % (gene,sense,self.__calcRPKM(gene,sense),anti_sense,self.__calcRPKM(gene,anti_sense),self.cDNA_count[gene]))
				else:
					file_out.write("%s\t%s\t%s\t%s\n" % (gene,sense,anti_sense,self.cDNA_count[gene]))
		self.sense_gene_hits = defaultdict(int)
		self.anti_sense_gene_hits = defaultdict(int)
		file_out.close()

	def __calcRPKM(self,gene,gene_hits):
		""" calculates the reads per kilobase per million mapped reads for a given gene
		Example:
		No.of mapped reads =3
		lenth of transcript=300 bp
		Total no. of reads =10,000
		rpk = 3/(300/1000) = 3/0.3 = 10
		rpkm = 10 / (10,000/1,000,000) = 10/ 0.01 = 1000
		"""
		coordinates = self.gtf.geneIterCoordinates(gene,coordinates=self.sequence,output="list")         # The length of the transcript is calculated using itercoordinates.
		try:
			gene_length = len([item for sublist in coordinates for item in sublist])                     # The reason for this is to make it possible to calculate FPKM/RPKM for features such as introns and exons
		except:
			gene_length = len(coordinates)
		rpk = float(gene_hits)/(gene_length/1000.0)
		rpkm = rpk /(float(self.mapped_reads)/1000000.0)
		return rpkm

	def __mutsAnnotationString(self,substitutions,deletions):
		"""Makes a CIGAR-type string for mutations in cDNAs"""
		mutlist = list()
		for pos in sorted(set(substitutions+deletions)):
			subsstring = str()
			delsstring = str()
			if pos in substitutions : subsstring = "S"
			if pos in deletions		: delsstring = "D"
			mutsstring = "%s%s%s" % (pos+1,subsstring,delsstring)	# add 1 to convert them to 1-based coordinates
			if subsstring or delsstring:
				mutlist.append(mutsstring)
		if mutlist:
			return ",".join(mutlist)
		else:
			return None

	def __setInsertVariables(self):
		"""does what it says"""
		self.__insert	= "reads"
		self.__insert2	= self.mapped_reads
		self.__insert3	= "reads"
		if self.duplicatesremoved:
			self.__insert	= "cDNAs"
			self.__insert3	= "reads"
		if self.readsareclustered:
			self.__insert	= "clusters"
			self.__insert3	= "cDNAs"
		if self.densities:
			self.__insert3	= "nucleotides"

	def __createFileHandles(self,file_type,file_extension="txt"):
		""" does what it says """
		filename = str()
		filename = "%s_%s_%s.%s" % (self.file_name,file_type,self.__insert,file_extension)
		self.file_list.append(filename)
		return filename

	def printIntronUTROverlapToGTF(self,min_overlap=1,out_file=None,extension="gtf"):
		"""Calculates how often a reads that were mapped to a gene overlapped with UTRs,intron or exons. Only does this for sense hits!"""
		assert self.gtf, "\n\nYou need to load a GTF annotation file using the readGTF method before you can use the printIntronUTROverlap method\n"
		assert self.hits_count, "\n\nYou need to run the countGeneHits method first before you can use the printIntronUTROverlap method.\n"
		self.__setInsertVariables()
		feature_counter = defaultdict(int)
		file_type		= "intron_and_UTR_overlap"
		file_extension	= "gtf"
		if out_file:
			self.file_name = out_file
		file_name = self.__createFileHandles(file_type,file_extension)
		file_out = NGSFileWriter(file_name)
		file_out.write("##gff-version 2\n")
		file_out.write("# generated by Counters version %s, %s\n" % (__version__,time.ctime()))
		file_out.write("# %s\n" %(' '.join(sys.argv)) )
		file_out.write("# total number of %s:\t%d\n" % (self.__insert,self.total_reads))
		file_out.write("# total number of mapped %s:\t%d\n" % (self.__insert,self.__insert2))
		file_out.write("# total number of paired reads:\t%d\n" % (self.paired_reads))
		file_out.write("# total number of single reads:\t%d\n" % (self.single_reads))
		file_out.write("# total number of overlapping genomic features\t%d\n" % (self.hits_count))
		file_out.write("# \tsense\t%d\n" % (self.sense_hits_count))
		file_out.write("# \tanti-sense\t%d\n" % (self.anti_sense_hits_count))
		file_out.write("# chromosome\tsource\tfeature\tstart\tend\t%s\tstrand\t.\tattributes\n" % (self.__insert3))
		for interval in sorted(self.count):		# sort by chromsome and start position
			chromosome,strand,start,end,substitutions,deletions = interval
			number_of_reads = self.count[interval]["coverage"]
			for gene in self.count[interval]["features"]:
				if gene != "no_matches":
					[five_UTR,three_UTR] = self.gtf.utrCoordinates(gene)
					if five_UTR:
						if overlap(start,end-1,five_UTR[0],five_UTR[1],overlap=min_overlap):
							mutsstring = self.__mutsAnnotationString(substitutions,deletions)
							feature_counter["5UTR"] += number_of_reads
							file_out.writeGTF(chromosome,self.__insert,"5UTR",start,end,score=number_of_reads,strand=self.gtf.strand(gene),gene_name=gene,gene_id=self.gtf.gene2orf(gene),comments=mutsstring)
					if three_UTR:
						if overlap(start,end-1,three_UTR[0],three_UTR[1],overlap=min_overlap):
							mutsstring = self.__mutsAnnotationString(substitutions,deletions)
							feature_counter["3UTR"] += number_of_reads
							file_out.writeGTF(chromosome,self.__insert,"3UTR",start,end,score=number_of_reads,strand=self.gtf.strand(gene),gene_name=gene,gene_id=self.gtf.gene2orf(gene),comments=mutsstring)
					introns = sorted(self.gtf.intronCoordinates(gene))
					if introns:
						for coord in introns:
							if overlap(start,end-1,coord[0],coord[1],overlap=min_overlap):
								mutsstring = self.__mutsAnnotationString(substitutions,deletions)
								feature_counter["intron"] += number_of_reads
								file_out.writeGTF(chromosome,self.__insert,"intron",start,end,score=number_of_reads,strand=self.gtf.strand(gene),gene_name=gene,gene_id=self.gtf.gene2orf(gene),comments=mutsstring)
		file_out.write("# total number:\n")
		file_out.write("#\t5'UTR:\t%d\n" % (feature_counter["5UTR"]))
		file_out.write("#\t3'UTR:\t%d\n" % (feature_counter["3UTR"]))
		file_out.write("#\tintrons:\t%d\n" % (feature_counter["intron"]))
		feature_counter = None		# empty the counter, no need to keep this into memory
		file_out.close()

	def printCountResultsToGTF(self,out_file=None,extension="gtf"):
		"""Generates a GTF output of the data analyses. If you want the gene names to be included in the GTF output file, add
		genes=True to the option. For this to work you do need to run the countGeneHits method first"""
		assert self.count, "\n\nNo reads found overlapping with genomic features. Potential reasons\n(1) The chromosome names in your GTF file are not identical to the chromosome names in your SAM/BAM or Novo file.\n(2) You have to run at least the countReads method first before you can print your results.\n"
		self.__setInsertVariables()
		file_type = "count_output"
		file_extension = extension
		if out_file:
			self.file_name = out_file
		file_name = self.__createFileHandles(file_type,file_extension)
		file_out = NGSFileWriter(file_name)
		file_out.write("##gff-version 2\n")
		file_out.write("# generated by Counters version %s, %s\n" % (__version__,time.ctime()))
		file_out.write("# %s\n" %(' '.join(sys.argv)) )
		file_out.write("# total number of %s:\t%d\n" % (self.__insert,self.total_reads))
		file_out.write("# total number of mapped %s:\t%d\n" % (self.__insert,self.__insert2))
		file_out.write("# total number of paired reads:\t%d\n" % (self.paired_reads))
		file_out.write("# total number of single reads:\t%d\n" % (self.single_reads))
		file_out.write("# total number of overlapping genomic features:\t%d\n" % (self.sense_hits_count+self.anti_sense_hits_count))
		file_out.write("#\tsense:\t%d\n" % (self.sense_hits_count))
		file_out.write("#\tanti-sense:\t%d\n" % (self.anti_sense_hits_count))
		file_out.write("# chromosome\tsource\tfeature\tstart\tend\t%s\tstrand\t.\tattributes\n" % (self.__insert3))
		for interval in sorted(self.count):
			gene_name = str()
			gene_id = str()
			chromosome,strand,read_start,read_end,substitutions,deletions = interval
			number_of_reads = self.count[interval]["coverage"]
			genes = self.count[interval]["features"]
			mutsstring = self.__mutsAnnotationString(substitutions,deletions)
			if genes:
				gene_name = ",".join(genes)
				gene_id = ",".join([self.gtf.gene2orf(gene) for gene in genes])
			else:
				gene_name = "no_matches"
				gene_id = "no_matches"
			file_out.writeGTF(chromosome,self.__insert,"interval",read_start,read_end,score=number_of_reads,strand=strand_converter(strand),gene_name=gene_name,gene_id=gene_id,comments=mutsstring)
		file_out.close()

	def printFileStats(self,out_file=None):
		"""Prints alignment file_stats to a file. User can supply its own output file name, which overrides default output"""
		assert self.count, "\n\nNo reads found overlapping with genomic features. Potential reasons\n(1) The chromosome names in your GTF file are not identical to the chromosome names in your SAM/BAM or Novo file.\n(2) You have to run at least the countReads method first before you can print your results.\n"
		self.__setInsertVariables()
		file_type = "file_statistics"
		if out_file:
			self.file_name = out_file
		file_name = self.__createFileHandles(file_type)
		file_out = open(file_name, "w")
		cumulative_hits = 0
		file_out.write("# generated by Counters version %s, %s\n" % (__version__,time.ctime()))
		file_out.write("# %s\n" %(' '.join(sys.argv)) )
		file_out.write("# total number of %s\t%d\n" % (self.__insert,self.total_reads))
		file_out.write("# total number of mapped %s:\t%d\n" % (self.__insert,self.__insert2))
		file_out.write("# total number of paired reads\t%d\n" % (self.paired_reads))
		file_out.write("# total number of single reads\t%d\n" % (self.single_reads))
		file_out.write("#\n# cDNA\tCumulative number of reads\n")
		self.file_list.append(file_name)
		for read_nr,key in enumerate(sorted(self.count),start=1):
			cumulative_hits += self.count[key]["coverage"]
			file_out.write("%d\t%d\n" % (read_nr,cumulative_hits))
		file_out.close()



#!/usr/bin/python

__author__ = "Sander Granneman"
__copyright__ = "Copyright 2018"
__version__ = "0.0.4"
__credits__ = ["Sander Granneman"]
__maintainer__ = "Sander Granneman"
__email__ = "sgrannem@staffmail.ed.ac.uk"
__status__ = "Production"

##################################################################################
#
#	fasta2dict.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import io
from collections import defaultdict

try:
    file_types = (file, io.IOBase)
except NameError:
    file_types = (io.IOBase,)

class Fasta2Dict():
	"""this class can be used to store the information in a fasta sequence file into a dictionary.
	The dictionary keys are the gene or chromosome names, whereas the "values" is a single string of sequence"""
	def __init__(self,fasta_file,ids=[]):
		self.__sequence = defaultdict(str)
		count = 0
		id = str()
		if type(fasta_file) == file_types:
		    fasta = fasta_file
		else:
		    fasta = open(fasta_file,"r")
		splitfasta = fasta.read().split('>')	 # This bit of code was inspired by the TAMO Fasta.py module
		for i in splitfasta[1:]:
			lines	 = i.split('\n')
			id		 = lines[0]
			sequence = ''.join(lines[1:])
			count	+= 1
			if not ids:
				self.__sequence[id] = sequence
			else:
				if id in ids: self.__sequence[id] = sequence
						
	def numberOfSeqs(self):
		return len(self.__sequence)
		
	def chromosomeLengths(self):
		for chr in sorted(self.__sequence):
			return "%s\t%s" % (chr,len(self.__sequence[chr]))
		
	def __getitem__(self,chromosome):
		return self.__sequence[chromosome]

	def __iter__(self):
		return iter(self.__sequence)

	def __setitem__(self,chromosome,sequence):
		self.__sequence[chromosome] = sequence

	def __call__(self):
		for key in self.__sequence:
			return "%s\t%s" % (key,self.__sequence[key])



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyClusterReads.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import time
from optparse import *
from collections import defaultdict
from pyCRAC.Parsers import GTF2
from pyCRAC.Methods import numpy_overlap
from pyCRAC.Classes.NGSFormatReaders import NGSFileReader
from pyCRAC.Classes import NGSFormatWriters
from pyCRAC.Classes.Clustering import *

def main(): 
	
	parser = OptionParser(usage="usage: %prog [options] -f filename", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("-f", "--input_file", dest="infile",help="provide the path to your GTF read data file. NOTE the file has to be correctly sorted! If you used pyReadCounters to generate the file you should be fine. If you modified it, use the sort command described in the manual to sort your file first by chromosome, then by strand and then by start position.", metavar="reads.gtf",default=None)
	files.add_option("-o", "--output_file", dest="outfile",help="provide a name for an output file. By default it writes to the standard output", metavar="clusters.gtf",default=None)
	files.add_option("--gtf",dest="annotation_file",metavar="Yourfavoritegtf.gtf",help="type the path to the gtf annotation file that you want to use",default=None)
	common = OptionGroup(parser,"Common pyCRAC options")
	common.add_option("-r","--range",dest="range",type="int",metavar="100",help="allows you to set the length of the UTR regions. If you set '-r 50' or '--range=50', then the program will set a fixed length (50 bp) regardless of whether the GTF annotation file has genes with annotated UTRs.",default=0)
	common.add_option("-a","--annotation",dest="annotation",help="select which annotation (i.e. protein_coding, ncRNA, sRNA, rRNA,snoRNA,snRNA, depending on the source of your GTF file) you would like to focus your analysis on. Default = all annotations",metavar="protein_coding",default=None)
	common.add_option("--debug",dest="debug",action="store_true",help=SUPPRESS_HELP,default=False)
	common.add_option("-v", "--verbose",action="store_true",help="to print status messages to a log file",default=False) 
	clusters = OptionGroup(parser, "Options for cluster analysis")
	clusters.add_option("--cic","--cdnasinclusters",dest="cdnasinclusters",metavar="2",type="int",help="sets the minimal number of overlapping cDNAs in each cluster. Default = 2",default=2)
	clusters.add_option("--co","--clusteroverlap", dest="clusteroverlap",metavar="5",action="store",type="int",help="sets the number of nucleotides cDNA sequences have to overlap to form a cluster. Default = 1 nucleotide",default=1)
	clusters.add_option("--ch","--clusterheight", dest="clusterheight",metavar="5",action="store",type="int",help="sets the minimal height of the cluster. Default = 2 nucleotides",default=2)
	clusters.add_option("--cl","--clusterlength", dest="clusterlength",metavar="100",action="store",type="int",help="to set the maximum cluster sequence length. Default = 200 nucleotides",default=200)		 
	clusters.add_option("--mutsfreq","--mutationfrequency",dest="mutsfreq",action="store",type="int",metavar="10",help="sets the minimal mutations frequency for a cluster position in the GTF output file. Default = 0%. Example: if the mutsfrequency is set at 10 and a cluster position has a mutated in less than 10% of the reads, then the mutation will not be reported.",default=0)
	parser.add_option_group(files)
	parser.add_option_group(common)
	parser.add_option_group(clusters)
	(options, args) = parser.parse_args()
	
	outfile = sys.stdout	### default output is standard output
	status	= sys.stdout	### status messages also go to the standard output
	logfile = open("pyClusterReads_log.txt","w")
	if not options.infile:
		parser.error("you forgot to include your GTF read data file\n")
	if options.verbose: status	= logfile
	if not options.outfile:
		status = logfile	# to prevent having status messages and data printed to the standard output.
	else:
		outfile = open(options.outfile,"w")
	if options.clusterlength < 100:
		parser.error("The minimal setting for the maximum cluster length flag is 100 nucleotides. Please change your settings.\n")
	
	list_of_tuples = defaultdict(list)
	
	### Making GTF2 parser object and parsing GTF annotation file ###
	status.write("Parsing GTF annotation data...\n")
	gtf = GTF2.Parse_GTF()
	gtf.read_GTF(options.annotation_file,source=options.annotation,ranges=options.range,transcripts=False)
	###

	### Making a gtf writer object ###
	gtfwriter = NGSFormatWriters.NGSFileWriter(options.outfile)
	headerstring = "##gff-version 2\n# generated by pyClusterReads.py version %s, %s\n# %s\n# chromosome\tfeature\tsource\tstart\tend\tmax_height\tstrand\tframe\tattributes\n" % (__version__,time.ctime(),' '.join(sys.argv))
	gtfwriter.write(headerstring)
	###

	### creating a ReadClustering object ###
	data = ReadClustering(options.infile,logfile)
	###
	
	### Iterating over the read data file ###
	mappedreads	  = 0
	sense_mapped  = 0
	anti_sense_mapped = 0
	overlappingfeats  = 0

	### finding clusters
	status.write("clustering intervals...\n")
	while True:
		iscluster = data.getCluster(min_overlap=options.clusteroverlap,min_cdnas=options.cdnasinclusters,max_length=options.clusterlength,min_height=options.clusterheight,debug=options.debug)
		if iscluster:
			if data.chromosome not in list_of_tuples:
				try:
					list_of_tuples[data.chromosome] = gtf.chromosomeGeneCoordIterator(data.chromosome,numpy=True)
				except AssertionError:
					continue
			search_results = numpy_overlap(list_of_tuples[data.chromosome],data.start,data.end-1,overlap=1)
			mutsstring = data.mutsAnnotationString(min_frequency=options.mutsfreq)
			if search_results:
				gene_name = ",".join(search_results)
				gene_id = ",".join([gtf.gene2orf(gene) for gene in search_results])
				overlappingfeats += len(search_results)
				for gene in search_results:
					if data.strand == gtf.strand(gene):
						sense_mapped += 1
					else:
						anti_sense_mapped += 1
			else:
				gene_name = "no_matches"
				gene_id = "no_matches"
			gtfwriter.writeGTF(data.chromosome,"cluster","interval",data.start,data.end,score=data.cluster_height,strand=data.strand,gene_name=gene_name,gene_id=gene_id,comments=mutsstring)
		else:
			break
	###				
	gtfwriter.write("# total number of intervals analysed:\t%d\n" % data.line_count)
	gtfwriter.write("# total number of clusters generated:\t%d\n" % data.cluster_count)
	gtfwriter.write("# total number of clusters overlapping genomic features:\t%d\n" % (sense_mapped+anti_sense_mapped))
	gtfwriter.write("# \tsense\t%d\n" % sense_mapped)
	gtfwriter.write("# \tanti-sense\t%d\n" % anti_sense_mapped)
	logfile.close()
	
if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__      = "Sander Granneman"
__copyright__   = "Copyright 2018"
__version__     = "0.3.3"
__credits__     = ["Sander Granneman"]
__maintainer__  = "Sander Granneman"
__email__       = "sgrannem@staffmail.ed.ac.uk"
__status__      = "Production"

##################################################################################
#
#	Novoalign.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
import six
import pprint
from pyCRAC.Methods import basequality,complement,reverse_complement
from pyCRAC.Classes.Exceptions import *

pp = pprint.PrettyPrinter(indent=4)

SUBS_REGEX		= re.compile('(\d+)([ATGC])>[ATGC]')
DELS_REGEX		= re.compile('(\d+)[-]([ATGC]+)')
INS_REGEX		= re.compile('(\d+)[+]([ATGC]+)')
COLLAPSER_REGEX = re.compile('>(\d+)-(\d+)_(.*)')

def reverse_strand(string):
	if not string:
		return None
	elif string	  == "F":
		return "R"
	elif string == "R":
		return "F"
	elif string == "+":
		return "-"
	else:
		return "+"

class Process_Novo():
	"""Base class for the Novoalign parser"""
	def __init__(self,file_in,base_quality=25):
		""" Initialises the class attributes"""
		self.headers		= list()
		self.__base_quality = base_quality
		self.__file			= None
		if hasattr(file_in,'read'):  					# This allows you to pipe data from the standard input. the input file would be sys.stdin
			self.__file = file_in
		elif isinstance(file_in,tuple):					# This allows you to process novoalign strings in lists
			self.__file = iter(file_in)
		else: 
			self.__file = open(file_in,"r")
		self.__line = str()
		self.__reset()

	def __reset(self):
		""" Resets all the attributes in the Novoalign class. This is done before it reads a new line in the data file. Surprisingly, this is much faster than __slots__ or using named tuples! Keeping it simple is usually better!"""
		self.__line			 = str()		# line of the current file
		self.__qname		 = str()		# Query sequence name
		self.__rname		 = str()		# Reference sequence name
		self.__strand_query	 = str()		# Indicates the strand of the query sequence, default is indicated
		self.__strand_mate	 = str()		# Indicates the strand of the mate sequence, default is indicated
		self.__ispaired		 = False		# Indicates whether the read is paired (True or False)
		self.__seqquery		 = str()		# Sequence
		self.__seqmate		 = str()		# Sequence of the mate sequence on the same strand as the reference
		self.__read_type	 = "S"			# Read is either the first read of the pair ("L") or the second/paired ("R"), default is indicated
		self.__qual			 = str()		# Query quality (ASCII-33=Phred base quality)
		self.__alignscore	 = int()		# Align score (Phred-scaled)
		self.__alignquality	 = int()		# Align quality (or mapping quality, Phred-scaled)
		self.__substitutions = list()		# List containing substitutions (0-based positions to reference sequence!)
		self.__insertions	 = list()		# List containing insertions	(0-based positions to reference sequence!)
		self.__deletions	 = list()		# List containing deletions		(0-based positions to reference sequence!)
		self.__skippedref	 = list()		# List containing numbers of nucleotides that were skipped from the reference
		self.__hit_repeat	 = "U"			# Indicates whether a read was mapped only once ("U") or has multiple alignment records ("R"), default is indicated
		self.__num_of_align	 = int()		# Indicates the number of alignment locations for a read on the genome
		self.__pos			 = int()		# 0-based leftmost mapping POSition
		self.__mpos			 = int()		# 0-based leftmost mapping POSition of the Mate read
		self.__mismatches	 = list()		# mismatches in the read
					
	def decode_Mismatches(self,mismatchstring):
		""" decodes the mismatch information at the end of each line. Novoalign mismatch coordinates are 1-based and are converted to 0-based coordinates """
		self.__substitutions = [(int(i[0])-1,i[1]) for i in SUBS_REGEX.findall(mismatchstring)]
		self.__deletions	 = [(int(i[0])-1,i[1]) for i in DELS_REGEX.findall(mismatchstring)]
		self.__insertions	 = [(int(i[0])-1,i[1]) for i in INS_REGEX.findall(mismatchstring)]
			
	def readline(self,correct_seq=True):
		""" Reads the next line in the file and keeps going until it finds a mapped read """
		self.__reset()
		while True:
			try:
				self.__line = six.next(self.__file)
			except StopIteration:
				return False
			if self.__line.startswith("#"):
				self.headers.append(self.__line)
				continue
			Fld =  self.__line.strip('\n').split('\t')
			if len(Fld) >=12:
				self.__qname		 = Fld[0]
				self.__read_type	 = Fld[1]
				self.__seqquery		 = Fld[2]
				self.__qual			 = Fld[3]
				self.__hit_repeat	 = Fld[4]
				self.__alignscore	 = int(Fld[5])
				self.__alignquality	 = int(Fld[6])
				self.__rname		 = Fld[7][1:]
				self.__pos			 = int(Fld[8])-1				# -1 converts position to 0-based coordinates!!
				self.__strand_query	 = Fld[9]
				self.__strand_mate	 = Fld[12]
				self.__mpos			 = int()
				if Fld[10] == ".": self.__ispaired = True
				if Fld[11] != ".": self.__mpos = int(Fld[11])-1		# -1 converts position to 0-based coordinates!!
				if self.__line[0] == ">":
					self.__qual = len(self.__seqquery)*"L"			# If there is no quality string, all bases get highest base quality value (Phred score of 41, assuming Illumina 1.8).
				try:
					self.__mismatches = Fld[13].split()
					self.decode_Mismatches(Fld[13])
				except IndexError:
					pass

#--------------------------------- Sequences are automatically corrected but this can be turned off! If reads contain deletions, dashes are inserted at the positions and substitutions are highlighted as lowercase letters ---------------------------------------------------
			
			if correct_seq: 
				self.correct_seq()
			return True
									
#-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
			
	def printHeaders(self):
		""" prints the Novo file header info """
		if self.headers:
			for line in self.headers:
				sys.stdout.write(line)
	
	def printDictionaryLayout(self):
		""" Prints the interval_dict dictionary layout using pretty printer"""
		pp.pprint(self.__dict__)
					
	def __call__(self):
		""" Returns the input line as a string"""
		if self.__line:
			return self.__line
		else:	
			return None
					
	def uniqueAlnLocation(self):
		""" From Novoalign documentation: column 5 in the novo file:
		U: A single alignment with this score was found. R: Multiple alignments with similar score were found.
		QC: The read was not aligned as it bases qualities were too low or it was a homopolymer read.
		NM :No alignment was found. QL: An alignment was found but it was below the quality threshold.\
		If a read has been aligned to only one genomic location (i.e. column 5 = "U") then
		this method will return 'True', otherwise it will return 'False'"""
		if self.__hit_repeat == "U":
			return True
		else:
			return False

	def seq_ID(self):
		""" Returns the sequence identification number of the read"""
		if self.__qname.startswith(">"):	# remove the chevron if it is before the header
			return self.__qname[1:]
		else:
			return self.__qname

	def read_type(self):
		""" Returns the read_type indicator ("S" = single end read; "L" or "R" indicate paired end reads)"""
		return self.__read_type
	
	def ismapped(self):
		""" Checks if a read is mapped to the reference sequence """
		if self.__rname:
			return True
		else:
			return False
	
	def sequence(self):
		""" Returns the sequence of the reads"""
		if self.__read_type == "R":
			return reverse_complement(self.__seqquery)
		else:		 
			return self.__seqquery

	def align_score(self):
		""" Returns the Phred format alignment score 10log10(P(R|Ai)).
		For status of 'R' and when not report alignment locations for repeats,
		the alignment score becomes the number of alignments to the read.
		For paired end the alignment score includes the fragment length penalty.
		Align_score is indicated in column 6 of the novo file."""
		if self.__alignscore:
			return self.__alignscore
		else:
			return 0

	def align_quality(self):
		""" Returns the align_quality value of the read sequence,
		which is the Phred format alignment quality score 10log10(1 P(Ai|R, G))
		using Sanger fastq coding method.Align quality is indicated in column 7 
		of the novo file"""
		if self.__alignquality:
			return self.__alignquality
		else:
			return 0

	def alignScoreThreshold(self,number):
		""" Checks whether the alignment score is smaller than the to a threshold set by the
		user. If the alignment score is smaller than the threshold, the method returns 'False' else it returns 'True' """
		if number > self.__alignscore:
			return False
		else:
			return True

	def alignQualityThreshold(self,number):
		""" Checks whether the alignment quality is smaller than to the to a threshold set by the
		user. If the alignment quality is smaller than the threshold, the method returns 'False' else it returns 'True' """
		if number > self.__alignquality:
			return False
		else:
			return True

	def lengthThreshold(self,number):
		""" This method tests whether the length of the read sequence is smaller or equal to a maximum length
		threshold set by the user. If the read is smaller than the threshold the method returns 'True', else
		it returns 'False' """
		if number >= len(self.__seqquery):
			return True
		else:
			return False

	def unique_hits(self):
		""" Returns R or U. R indicates that the read was mapped to multiple genomic locations.
		U indicates that the read mapped to only one location on the genome"""
		return self.__hit_repeat

	def sequenceAlignmentLocations(self):
		"""This method returns the number of alignments to the read"""
		if self.__hit_repeat == "R":
			return self.__alignscore
		if self.__hit_repeat == "U":
			return 1

	def read_start(self):
		""" Returns the 0-based start position of the read on the chromosome"""
		if self.__pos >= 0:
			return self.__pos
		else:
			return None

	def read_start_paired(self):
		""" Returns the start position of the paired read on the chromosome"""
		if self.__mpos >= 0:
			return self.__mpos
		else:
			return None
					
	def paired_sequence(self,distance=1000):
		""" Returns TRUE if the read has a paired read mapped to the same chromosomal location"""
		if self.__ispaired and self.__pos - self.__mpos <= distance or self.__pos - self.__mpos >= distance:
			return True
		else:
			return None
					
	def read_end_paired(self):
		""" Returns the chromosomal end coordinate of the paired read. For paired end reads
		it calculates the end position using the information obtained from the paired read."""
		if self.__mpos <= self.__pos:
			return self.__pos + len(self.__seqquery) - 1
		else:
			return None

	def read_end(self):
		""" Returns the chromosomal end coordinate of the read. For paired end reads
		it calculates the end position using the information obtained from the paired read."""
		if self.__pos >= 0:
			return self.__pos + len(self.__seqquery)
		else:
			return None

	def strand(self):
		""" Returns the read strand of the chromosome ("F" = watson, "R" = crick strand) the read was mapped to
		NOTE: if the read was obtained from the reverse sequencing reaction (i.e. read_type = "R"), then the opposite strand
		will be returned"""
		if self.__read_type == "R":
			return reverse_strand(self.__strand_query)
		else:
			return self.__strand_query

	def strand_paired(self):
		""" Returns the strand of the paired read"""
		if self.__strand_mate:
			return self.__strand_mate
		else:
			return None
	
	def chr_ID(self):
		""" Returns the reference chromosome name/number the read sequence was mapped to"""
		if self.__rname:
			return self.__rname
		else:
			return None

	def baseQuality(self,nucl_nr,qualtype="L"):
		""" This method returns the basequality value (an integer, NOT ASCI!) at position nucl_nr. Usage: base_quality(24)
		NOTE: This method can ONLY be used when the sequence has been corrected for deletions and insertions! (use the correct_seq
		or genomic_seq methods to correct the sequence. You need to include format of the quality scoring, for example: qualtype="sanger" or qualtype="illumina
		S - Sanger		  Phred+33,	 raw reads typically (0, 40)
		X - Solexa		  Solexa+64, raw reads typically (-5, 40)
		I - Illumina 1.3+ Phred+64,	 raw reads typically (0, 40)
		J - Illumina 1.5+ Phred+64,	 raw reads typically (3, 40) with 0=unused, 1=unused, 2=Read Segment Quality Control Indicator (bold) 
		L - Illumina 1.8+ Phred+33,	 raw reads typically (0, 41)
		Default is L """
		quality = self.baseQualityString()
		if len(quality) > 1:
			if nucl_nr > len(quality):
				corr_nr = nucl_nr - self.__pos
				if corr_nr < 0:
					raise BaseQualityError("\n\ncould not calculate the base-quality for nucleotide position %d in read %s\nQuality length\t%s\nposition\t%s" % (nucl_nr,self.__qname,len(quality),nucl_nr))
				else:
					return basequality(quality[corr_nr],ftype=qualtype)
			else:
				return basequality(quality[nucl_nr],ftype=qualtype)
		else:
			return 0
							
	def baseQualityString(self):
		""" Returns the quality string for the sequence in the line"""
		if self.__read_type == "S" or self.__read_type == "L":
			if self.__strand_query == "F":
				return self.__qual
			else:
				return self.__qual[::-1]
		elif self.__read_type == "R":
			if self.__strand_query == "F":
				return self.__qual[::-1]
			else:
				return self.__qual
							
	def mutations(self):
		""" This method returns a list of all the mutations (basically column 13 in the .novo file) found in a read sequence"""
		if self.__mismatches:
			return self.__mismatches
		else:
			return []

	def deletions(self,chromosome_location=False):
		""" Returns a list of deletions found in a read. Offset of mismatches are 0-based relative 
		to the reference sequence alignment location. They are not the location of the mismatches in the read itself
		but in the reference sequence!! If the values are negative then the read is the reverse complement of the reference sequence. 
		If you include the argument 'chromosome_location = True' then the method will only return the chromosomal coordinate. """
		if chromosome_location and self.__deletions:
			return [i[0]+self.__pos for i in self.__deletions]		 # returns a 0-based position			 
		elif self.__deletions:
			if self.__strand_query == "F":
				return [i[0] for i in self.__deletions]	  
			elif self.__strand_query == "R":
				return [-(i[0]+1) for i in self.__deletions]	
		else:
			return [] 
	
	def substitutions(self,chromosome_location=False):
		""" Returns a list of substitutions found in a read. Offset of mismatches are 0-based relative 
		to the reference sequence alignment location. If the values are negative then the read is the reverse complement of the reference sequence.
		If you include the argument 'chromosome_location = True' then the method will return the location of the substitutions on the chromosome.
		If you add basequality=30 then all the substitutions with a basequality lower than 30 will be removed. If you set filtersubs="TC" then
		only T to C conversions will be kept and others will be removed.
		All are 0-based positions"""
		if self.__substitutions:
			if chromosome_location and self.__substitutions:
				return [i[0]+self.__pos for i in self.__substitutions]	  # returns a 0-based position			 
			elif self.__substitutions:
				if self.__strand_query	 == "F":
					return [i[0] for i in self.__substitutions]	  
				elif self.__strand_query == "R":
					return [-(i[0]+1) for i in self.__substitutions]	
			else:
				return []
		else:
			return []

	def filterMutations(self,filter):
		"""Checks if a read has specific deletions or substitutions:
		Choices are:
				'delsonly': checks if a read has deletions.		 If so, it returns True and only deletions will be reported. Else, it returns False.
				'subsonly': checks if a read has substitutions.	 If so, it returns True and only substitutions will be reported. Else, it returns False.
				'nomuts'  : no mutations will be reported.
				'TC'	  : checks if a read has T->C mutations. If so, it removes mutations that do not fit these criteria and returns True, else False.
				'[T,C,G]' : (list) checks if the reads has T's, C's and/or G's mutated. If so, it removes mutations that do not fit these criteria and returns True, else False.
		"""
		if filter	== "delsonly":
			self.__substitutions = []
			if self.__deletions: 
				return True
			else: 
				return False
		elif filter == "subsonly" and self.__substitutions:
			self.__deletions = []
			if self.__substitutions: 
				return True
			else: 
				return False
		elif filter == "nomuts":
			self.__substitutions = []
			self.__deletions	 = []
			return True
				
		elif isinstance(filter,list):
			# format for deletions: 'offset'>'refbase'
			# format for substitutions: 'offset''refbase'>'readbase'
			if self.__substitutions:
				newmutslist = list()
				for (position,nucleotide) in self.__substitutions:
					if self.__read_type == "S" or self.__read_type == "L":
						if self.__strand_query == "R":
							nucleotide = reverse_complement(nucleotide)
					elif self.__read_type == "R":
						if self.__strand_query == "F":
							nucleotide = reverse_complement(nucleotide) 
					if nucleotide.upper() in filter:
						newmutslist.append((position,nucleotide))				# keep only those substitutions that match your criteria
				if newmutslist: 
					self.__substitutions = newmutslist
					return True
				else:
					return False
					
			if self.__deletions:
				newmutslist = list()
				for (position,nucleotide) in self.__deletions:
					if self.__read_type == "S" or self.__read_type == "L":
						if self.__strand_query == "R":
							nucleotide = reverse_complement(nucleotide)
					elif self.__read_type == "R":
						if self.__strand_query == "F":
							nucleotide = reverse_complement(nucleotide) 
					if nucleotide.upper() in filter:
						newmutslist.append((position,nucleotide))				# keep only those deletions that match your criteria
				if newmutslist: 
					self.__deletions = newmutslist	  
					return True
				else:
					return False	
			
			mutations = self.__substitutions + self.__deletions
			if mutations:
				return True						   
			else:
				return False

		elif isinstance(filter,str) and len(filter) == 2:
			# format = 'offset''refbase'>'readbase' 
			self.__deletions = []
			if self.__substitutions:	
				newmutslist = list()
				for subs in self.__substitutions:
					(position,nucleotide) = subs
					mutation = self.sequence()[position]		# read base
					actual	 = nucleotide						# ref base
					if self.__read_type == "S" or self.__read_type == "L":
						if self.__strand_query == "R":
							mutation = self.sequence()[-(position+1)]		# read base
							actual	 = reverse_complement(nucleotide)		# ref base
					elif self.__read_type == "R":
						if self.__strand_query == "F":
							mutation = self.sequence()[-(position+1)]		# read base
							actual	 = reverse_complement(nucleotide)		# ref base
					string = "%s%s" % (actual.upper(),mutation.upper())
					if string == filter:
						newmutslist.append(subs)			# keep only those substitutions that match your criteria
				self.__substitutions = newmutslist
				if self.__substitutions:
						return True
				else:
						return False
			else:
				return False
		else:
			return False
							
	def subsBaseQualFilter(self,basequality=0):
		""" checks the base quality of substitutions in the substitutions list. Returns a filtered list. The method
		expects illumina 1.8+ values"""
		if self.__substitutions:
			for subs in sorted(self.__substitutions):
				if self.baseQuality(subs[0]) < basequality:
					self.__substitutions.remove(subs)
						
		else:
			return []
				 
	def seq_length(self):
		""" Returns the length of the read sequence"""
		return len(self.__seqquery)
			
	def correct_seq(self):
		""" Checks if the sequence has any mutations and if so then removes the insertions
		adds dashes where it found a deletions and highlights substitutions lowercase. This
		method also corrects the base quality string (3rd column in novo file) if deletions
		are present in the sequence. This is essential if you want to know base quality values
		of substitutions"""
		nucleotides = list()
		quality		= list(self.__qual)
		if self.__qname and self.__mismatches:
			try:
				if self.__strand_query == "R":
					nucleotides = list(reverse_complement(self.__seqquery))
					quality.reverse()
				else:
					nucleotides = list(self.__seqquery)
				for i in self.__insertions:
					position,insertion = i[0],i[1]
					length			   = len(insertion)
					del nucleotides[position:position+length]
					del quality[position:position+length]
				for i in self.__deletions:
					position,deletions = i[0],i[1]
					for i in range(len(deletions)):
						nucleotides.insert(position,"-")
						quality.insert(position,"!")	# ! is base_quality of 0
				for i in self.__substitutions:
					position,substitution = i[0],i[1]
					nucleotides[position] = nucleotides[position].lower()
					quality[position]	  = quality[position].lower()
				if self.__strand_query == "R":
					self.__seqquery = reverse_complement(''.join(nucleotides))
					quality.reverse()
					self.__qual		= ''.join(quality)
				else:
					self.__qual		= ''.join(quality)
					self.__seqquery = ''.join(nucleotides)
				return True
			except:
				sys.stderr.write("Error in line %s\tCould not correct the sequence. Please remove from file\n" % self.__line.strip())
				return False
		else:
			return False
		
	def genomic_seq(self):
		""" Checks if the sequence has any mutations and if so then removes the insertions
		corrects the deletions and substitutions and shows these in lowercase. This method
		basically corrects the read sequence to the genomic sequence"""
		nucleotides = list()
		sequence	= str()
		if self.__mismatches:
			try:
				if self.__strand_query == "R":
					nucleotides = list(reverse_complement(self.__seqquery))
				else:
					nucleotides = list(self.__seqquery)
				for i in self.__deletions:
					position,deletions = i[0],i[1]
					for i in range(len(deletions)):
						nucleotides[position] = deletions[i].lower()
				for i in self.__substitutions:
					position,substitution = i[0],i[1]
					nucleotides[position] = substitution.lower()
				if self.__strand_query == "R":
					sequence = reverse_complement(''.join(nucleotides))
				elif self.__read_type == "R":
					sequence = reverse_complement(''.join(nucleotides))
				else:
					sequence = ''.join(nucleotides)
				return sequence
			except:
				sys.stderr.write("Error in line %s\tCould not generate the genomic sequence for this read. Please remove from file\n" % self.__line.strip())
				return self.__seqquery
		else:
			return self.__seqquery

	def fasta(self):
		""" Returns the read ID and read sequence in the fasta format"""
		if self.__qname:
			fasta =	 ">%s\n%s" % (self.__qname,self.__seqquery)
			return fasta
	
	def tab(self):
		""" Returns the read ID and read sequence in the tab delimited format"""
		if self.__qname:
			tab =  "%s\t%s" % (self.__qname,self.__seqquery)
			return tab



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2020"
__version__		= "0.1.2"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyCalculateFDRs.py
#
#
#	Copyright (c) Sander Granneman 2020
#
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import os
import time
import shutil
from optparse import *
from pyCRAC.Classes.NGSFormatWriters import NGSFileWriter
from pyCRAC.Classes.FDR import *
from pyCRAC.Parsers import GTF2
from pyCRAC.Methods import splitData,processChromFile,contigousarray2Intervals,getfilename

def main():
	parser = OptionParser(usage="usage: %prog [options] -f readdatafile --gtf=GTF_annotation_file -o output_file.gtf",version="%s" % __version__)
	parser.add_option("-f", "--readdatafile", dest="readdatafile",help="Name of the bed/gff/gtf file containing the experiment read/cDNA coordinates", metavar="experiment_file",default=None)
	parser.add_option("--control", dest="controldatafile",help="Name of the bed/gff/gtf file containing the control read/cDNA coordinates",metavar="control_file",default=None) # help = Name of the bed/gff/gtf file containing the control data read/cDNA coordinates. If no control data is provided, the reads from the experimental data will be randomely shuffled over genomic features
	parser.add_option("--file_type",dest="file_type",choices=["bed","gtf","gff"],help="this tool supports bed6, gtf and gff input files. Please select from 'bed','gtf' or 'gff'. Default=gtf",default="gtf")
	parser.add_option("-o", "--outfile",dest="outfile",help="Optional. Provide the name of the output file. Default is 'selected_intervals.gtf'",metavar="outfile.gtf",default="interval_FDRs.gtf")
	parser.add_option("-l","--logfile",dest="logfile",help="Optional. With this flag the program prints a log file that contains information about the statistical analyses. Default is none",default=False)
	parser.add_option("-r","--range",dest="range",type="int",metavar="100",help="allows you to set the length of the UTR regions. If you set '-r 50' or '--range=50', then the program will set a fixed length (50 bp) regardless of whether the GTF file has genes with annotated UTRs.",default=False)
	parser.add_option("-a","--annotation",dest="annotation",help="select which annotation (i.e. protein_coding, ncRNA, sRNA, rRNA,snoRNA,snRNA, depending on the source of your GTF file) you would like to focus your analysis on. Default = all annotations",metavar="protein_coding",default=None)
	parser.add_option("-s","--sequence",dest="sequence",choices=["genomic","coding"],metavar="genomic",help="use 'coding' if you suspect your protein only binds to spliced mRNAs. Default = genomic",default="genomic")
	parser.add_option("-c", "--chromfile",dest="chromfile",help="Location of the chromosome info file. This file should have two columns: first column is the names of the chromosomes, second column is length of the chromosomes. Default is yeast",metavar="yeast.txt",default=None)
	parser.add_option("--gtf",dest="annotation_file",help="Name of the annotation file. Default is %s" % None, metavar="yeast.gtf",default=None)
	parser.add_option("-v", "--verbose",action="store_true",help="to print status messages to a log file",default=False)
	parser.add_option("-m","--minfdr",dest="minfdr",type="float",help="To set a minimal FDR threshold for filtering interval data. Default is 0.05",default=0.05)
	parser.add_option("--min",dest="min",help="to set a minimal read coverages for a region. Regions with coverage less than minimum will be ignoredve an FDR of zero",type="int",default=1)
	parser.add_option("--iterations",dest="iterations",type="int",help="to set the number of iterations for randomization of read coordinates. Default=100",default=100)
	parser.add_option("--debug",dest="debug",action="store_true",help=SUPPRESS_HELP,default=False)
	parser.add_option("--todisk",dest="todisk",action="store_true",help=SUPPRESS_HELP,default=False)
	parser.add_option("--featlengthlimit",dest="featlength",type="int",help=SUPPRESS_HELP,default=5000000)
	(options, args) = parser.parse_args()
	debug	= options.debug
	status	= sys.stdout
	if options.verbose: status = open("pyCalculateFDRs_log.txt","w")

	### Making GTF2 parser object and parsing GTF annotation file ###
	status.write("Parsing GTF annotation data...\n")
	gtf = GTF2.Parse_GTF()
	gtf.read_GTF(options.annotation_file,ranges=options.range,source=options.annotation,transcripts=False)
	###

	list_of_tuples = defaultdict(list)
	experimentdataarray = defaultdict(list)
	controldataarray	= defaultdict(list)

	### Creating outfile, logfile and outfile GTF writer object ###
	outfile = NGSFileWriter(options.outfile)
	if options.logfile:
		logfile = open("%s_features_FDR_data.txt" % getfilename(options.outfile),"w")
	###

	### Making a temporary directory
	tempdir = tempfile.mkdtemp(prefix="pyCalculateFDRs_",suffix="_temp",dir="./")
	controltempdir = None

	###

	### It is best to split up the data file in temp files, one for each chromosome ###
	status.write("Creating temporary files for each chromosome...\n")
	status.write("Processing the experimental data...\n")
	datafiles = splitData(options.readdatafile,tempdir)

	### if there is a control dataset, do the same:

	if options.controldatafile:
		status.write("Processing the control data...\n")
		controltempdir = tempfile.mkdtemp(prefix="pyCalculateFDRs_",suffix="_temp",dir="./")
		controlfiles   = splitData(options.controldatafile,controltempdir)

	###

	### Creating a ModFDR object ###
	fdr = ModFDR(iterations=options.iterations,mincoverage=options.min,debug=options.debug,minfdr=options.minfdr)
	###

	chromlengthfile = open(options.chromfile,"r").readlines()
	chromdict = processChromFile(chromlengthfile)

	### Making a gtf writer object ###
	gffversionstring = "##gff-version 2\n"
	headerstring = "# generated by pyCalculateFDRs version %s, %s\n# %s\n" % (__version__,time.ctime(),' '.join(sys.argv))
	gtfstring	 = "# chromosome\tfeature\tsource\tstart\tend\tpeak_height\tstrand\tFDR\tattributes\n"

	outfile.write("%s%s%s" % (gffversionstring,headerstring,gtfstring))
	if options.logfile:
		logfile.write(headerstring)

	### iterating through the chromosome data files
	for chromosome in sorted(datafiles):
		status.write("Processing data for chromosome %s...\n" % chromosome)
		try:
			chromosomelength = chromdict[chromosome]
			chromosomelength += (2* options.range)	   # otherwise the program would throw an index error for genes close to the edges of the chromosome
		except KeyError:
			sys.stderr.write("\nWARNING! Could not find chromosome %s in the chromosome info file. Please double check if the chromosome names of the chromosome info file and read GTF file match\n\n")
			continue
		file = datafiles[chromosome]
		experimentdataarray,genecoverageinfo = fdr.processReadData(gtf,file.name,chromosomelength,todisk=options.todisk,tempdirectory=tempdir)	# doing this one chromosome at a time significantly reduces memory usage. The datafile does not have to be sorted.
		if options.controldatafile:
			if chromosome in controlfiles:
				file = controlfiles[chromosome]
				controldataarray,controlcoverageinfo = fdr.processReadData(gtf,file.name,chromosomelength,todisk=options.todisk,tempdirectory=controltempdir)
			else:
				controldataarray["+"] = np.zeros(chromosomelength)
				controldataarray["-"] = np.zeros(chromosomelength)
		if options.logfile:
			logfile.write("\n####### chromosome %s #######\n" % chromosome)
		for gene in genecoverageinfo:
			itercoordinates = gtf.geneIterCoordinates(gene,coordinates=options.sequence)
			strand = gtf.strand(gene)
			gene_length = len(itercoordinates)
			if gene_length > options.featlength:
				sys.stderr.write("\tgene %s is longer than the %s nucleotide feature limit and will be ignored\n" % (gene,options.featlength))
				continue
			actualdata = experimentdataarray[strand][itercoordinates]
			genestart = itercoordinates[0]
			if actualdata.any():
				if options.controldatafile:
					controldata	  = controldataarray[strand][itercoordinates]
					signifregions = fdr.findSignificantRegions(actualdata,controldata,randomised=False)
				else:
					controldata  = fdr.shuffleReadPositions(gene_length,genecoverageinfo[gene])
					signifregions = fdr.findSignificantRegions(actualdata,controldata)

				if signifregions.any():		# if there are any significant regions...
					if options.logfile:
						logfile.write("\n# %s\nheight\tFDR\texperimental_data_hits\tcontrol_data_hits\n" % gene)
						fdr.writeFDRData(logfile)
					for (first,second) in contigousarray2Intervals(signifregions):
						start = genestart + first
						end = genestart + second + 1
						coordinates = np.arange(start,end)
						peak_height = experimentdataarray[strand][coordinates].max()
						peak_height_fdr = fdr.calculatePeakHeightFDR(actualdata,controldata,peak_height)
						outfile.writeGTF(chromosome,\
										"cluster",\
										"interval",\
										start,\
										end,\
										score=peak_height,\
										strand=strand,\
										#frame=f"{peak_height_fdr:.4f}",\ ### Python 3 code
										frame="%.4f" % peak_height_fdr,\
										gene_name=gene,\
										gene_id=gtf.gene2orf(gene),\
										comments=None)
		if options.todisk:
			for i in chromarray["files"]:
				os.unlink(i.name)
		os.unlink(file.name)	# delete the temporary file
	status.write("Removing remaining temporary files...\n")
	try:
		shutil.rmtree(tempdir)
		if controltempdir:
			shutil.rmtree(controltempdir)
	except:
		status.write("\nERROR! Could not delete the temporary files because of permission issues, please do this manually\n")

if __name__ == "__main__":
	main()
